unit RakCanvasData;

{///////////////////////////////////////////////////////////////////////////////
// Beschreibung : Die Unit RakCanvasData stellt die Klasse TRakCanvasData
                  zur Verfgung.
                  TRakCanvasData untersttzt die Umrechnung zwischen
                  gerteunabhngigen und gerteabhngigen Koordinaten
// erstellt am  : ???
////////////////////////////////////////////////////////////////////////////////
// nderungen:
                13.10.01 Bugfix in TRakCanvasData.MmRectToPixelRect
///////////////////////////////////////////////////////////////////////////////}
// Delphi 5 Implementation
//
// Copyright (c) 2000 by Richard Kaspar
// ("http://www.kasparsoft.de")
//
// Definitions:
// - "Package" refers to the collection of files distributed by
//   the Copyright Holder, and derivatives of that collection of
//   files created through textual modification.
// - "Standard Version" refers to such a Package if it has not
//   been modified, or has been modified in accordance with the
//   wishes of the Copyright Holder.
// - "Copyright Holder" is whoever is name in the copyright or
//   copyrights for the package.
// - "You" is you, if you're thinking about copying or distributing
//   this Package.
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Package"), to deal in the Package without restriction,
// including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the
// Package, and to permit persons to whom the Package is furnished
// to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Package.
//
// You may modify your copy of this Package in any way, provided
// that you insert a prominent notice in each changed file stating
// how and when you changed a file, and provided that you do at
// least one of the following:
//
// a) allow the Copyright Holder to include your modifications in
// the Standard Version of the Package.
//
// b) use the modified Package only within your corporation or
// organization.
//
// c) rename any non standard executables, units, and classes so
// the names do not conflict with standard executables, units, and
// classes, and provide a separate manual page that clearly documents
// how it differs from the standard version.
//
// d) make other distribution arrangements with the Copyright Holder.
//
// The name of the Copyright Holder may not be used to endorse or
// promote products derived from this Package without specific prior
// written permission.
//
// THE PACKAGE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// PACKAGE OR THE USE OR OTHER DEALINGS IN THE PACKAGE.
{$I RAK.INC}

interface

uses  Windows, Graphics,
  RakObject, misctype, miscfunc;

type
{////////////////////////////////////////////////////////////////////////////////
// Klasse       :
// Beschreibung : TRakCanvasData untersttzt die Umrechnung zwischen
                  gerteunabhngigen und gerteabhngigen Koordinaten.
                  TRakCanvasData geht davon aus, dass die
                  gerteunabhngigen Koordinaten in Millimeter angegeben sind

                  TRakCanvasData wird innerhalb des Open-Source-Codes meines
                  allgemeinen grafischen Editors massiv verwendet.
                  Siehe Unit RakGraphicObject
// erstellt am  : ???
////////////////////////////////////////////////////////////////////////////////
// nderungen:
  16.4.2002 Property Aufloesung ist jetzt auch lesbar
////////////////////////////////////////////////////////////////////////////////}
  TRakCanvasData = class(TRakObject)
  private
    FAufloesungY: extended;
    FAufloesungX: extended;
    FMmOffsetX: extended;
    FMmOffsetY: extended;

    procedure SetAufloesungX(const Value: extended);
    procedure SetAufloesungY(const Value: extended);
    procedure SetMmOffsetX(const Value: extended);
    procedure SetMmOffsetY(const Value: extended);
    procedure SetAufloesung(const Value: extended);
    function GetAufloesung: extended;
  protected
    procedure CreateNow; override;
  public
    procedure Assign(Source: TRakCanvasData);

    // Rechnet einen Millimeter-Wert der X-Achse in Pixel um
    function MmToPixelX(Mm: extended; UseOffset: boolean = false;
      Grenze: real = 0.5): integer;
    // Rechnet einen Millimeter-Wert der Y-Achse in Pixel um
    function MmToPixelY(Mm: extended; UseOffset: boolean = false;
      Grenze: real = 0.5): integer;
    // Rechnet ein Millimeter-Rechteck in ein Pixel-Rechteck um
    function MmRectToPixelRect(ARect: TExtRect; UseOffset: boolean = false;
      Grenze: real = 0.5): TRect;
    // Rechnet ein Millimeter-Polygon in ein Pixel-Polygon um
    function MmPolygonToPixelPolygon(APolygon: TExtPolygon; UseOffset: boolean = true;
      Grenze: real = 0.5): TPolygon;
    // Rechnet einen Millimeter-Punkt in einen Pixel-Punkt um
    function MmPointToPixelPoint(APoint: TExtPoint; UseOffset: boolean = true;
      Grenze: real = 0.5): TPoint;

    // Rechnet einen Pixel-Wert der X-Achse in Millimeter um
    function PixelToMmX(Pixel: integer; UseOffset: boolean = false): extended;
    // Rechnet einen Pixel-Wert der Y-Achse in Millimeter um
    function PixelToMmY(Pixel: integer; UseOffset: boolean = false): extended;
    // Rechnet ein Pixel-Rechteck in ein Millimeter-Rechteck um
    function PixelRectToMmRect(ARect: TRect; UseOffset: boolean = false): TExtRect;
    // Rechnet ein Pixel-Polygon in ein Millimeter-Polygon um
    function PixelPolygonToMmPolygon(APolygon: TPolygon; UseOffset: boolean = true): TExtPolygon;
    // Rechnet einen Pixel-Punkt in einen Millimeter-Punkt um
    function PixelPointToMmPoint(APoint: TPoint; UseOffset: boolean = true): TExtPoint;

    // Mit Aufloesung kann man die beiden Properties AufloesungX und AufloesungY auf einmal setzen
    // Wenn man lesend darauf zugreift wird der absolute Mittelwert von AufloesungX und AufloesungY zurckgegeben
    property Aufloesung: extended read GetAufloesung write SetAufloesung;
    // AufloesungX gibt die Pixel pro Millimeter in der X-Achse an
    property AufloesungX: extended read FAufloesungX write SetAufloesungX;
    // AufloesungY gibt die Pixel pro Millimeter in der Y-Achse an
    property AufloesungY: extended read FAufloesungY write SetAufloesungY;
    // MmOffsetX legt den Nullpunkt des Koordinatensystems in der X-Achse fest
    property MmOffsetX: extended read FMmOffsetX write SetMmOffsetX;
    // MmOffsetY legt den Nullpunkt des Koordinatensystems in der Y-Achse fest
    property MmOffsetY: extended read FMmOffsetY write SetMmOffsetY;
  end;

procedure PaintWithCanvasData(cd: TRakCanvasData; Canvas: TCanvas; APolygon: TExtPolygon); overload;

implementation


procedure PaintWithCanvasData(cd: TRakCanvasData; Canvas: TCanvas; APolygon: TExtPolygon);
begin
  Canvas.Polygon(cd.MmPolygonToPixelPolygon(APolygon));
end;


{ TRakCanvasData }

procedure TRakCanvasData.Assign(Source: TRakCanvasData);
begin
  FMmOffsetX := Source.FMmOffsetX;
  FAufloesungY := Source.FAufloesungY;
  FAufloesungX := Source.FAufloesungX;
  FMmOffsetY := Source.FMmOffsetY;
end;

procedure TRakCanvasData.CreateNow;
begin
  inherited;

  FMmOffsetX := 0;
  FMmOffsetY := 0;
  FAufloesungY := 10;
  FAufloesungX := 10;
end;

function TRakCanvasData.GetAufloesung: extended;
begin
  Result := abs((AufloesungX + AufloesungY) /2);
end;

function TRakCanvasData.MmPointToPixelPoint(APoint: TExtPoint;
  UseOffset: boolean; Grenze: real): TPoint;
begin
  Result.X := MmToPixelX(APoint.X, UseOffset, Grenze);
  Result.Y := MmToPixelY(APoint.Y, UseOffset, Grenze);
end;

function TRakCanvasData.MmPolygonToPixelPolygon(APolygon: TExtPolygon;
  UseOffset: boolean; Grenze: real): TPolygon;
var
  i: integer;
begin
  SetLength(Result, Length(APolygon));
  for i := 0 to Length(APolygon) - 1 do
    Result[i] := MmPointToPixelPoint(APolygon[i], UseOffset, Grenze);
end;

function TRakCanvasData.MmRectToPixelRect(ARect: TExtRect;
  UseOffset: boolean; Grenze: real): TRect;
begin
  Result.Left := MmToPixelX(ARect.Left, UseOffset, Grenze);
  Result.Top := MmToPixelY(ARect.Top, UseOffset, Grenze);
  Result.Right := MmToPixelX(ARect.Right, UseOffset, Grenze);
  Result.Bottom := MmToPixelY(ARect.Bottom, UseOffset, Grenze);
  if (Result.Left = Result.Right) and (ARect.Left < ARect.Right) then
    inc(Result.Right);
  if (Result.Top = Result.Bottom) and (ARect.Top < ARect.Bottom) then
    inc(Result.Bottom);
end;

function TRakCanvasData.MmToPixelX(Mm: extended;
  UseOffset: boolean; Grenze: real): integer;
begin
  if UseOffset then
    Result := RoundInt((Mm + MmOffsetX) * AufloesungX, Grenze)
  else
    Result := RoundInt(Mm * AufloesungX, Grenze);
end;

function TRakCanvasData.MmToPixelY(Mm: extended;
  UseOffset: boolean; Grenze: real): integer;
begin
  if UseOffset then
    Result := RoundInt((Mm + MmOffsetY) * AufloesungY, Grenze)
  else
    Result := RoundInt(Mm * AufloesungY, Grenze);
end;

function TRakCanvasData.PixelPointToMmPoint(APoint: TPoint;
  UseOffset: boolean): TExtPoint;
begin
  Result.X := PixelToMmX(APoint.X, UseOffset);
  Result.Y := PixelToMmY(APoint.Y, UseOffset);
end;

function TRakCanvasData.PixelPolygonToMmPolygon(APolygon: TPolygon;
  UseOffset: boolean): TExtPolygon;
var
  i: integer;
begin
  SetLength(Result, Length(APolygon));
  for i := 0 to Length(APolygon) - 1 do
    Result[i] := PixelPointToMmPoint(APolygon[i], UseOffset);
end;

function TRakCanvasData.PixelRectToMmRect(ARect: TRect; UseOffset: boolean): TExtRect;
begin
  Result.Left := PixelToMmX(ARect.Left, UseOffset);
  Result.Top := PixelToMmY(ARect.Top, UseOffset);
  Result.Right := PixelToMmX(ARect.Right, UseOffset);
  Result.Bottom := PixelToMmY(ARect.Bottom, UseOffset);
end;

function TRakCanvasData.PixelToMmX(Pixel: integer;
  UseOffset: boolean): extended;
begin
  if UseOffset then
    Result := Pixel / AufloesungX - MmOffsetX
  else
    Result := Pixel / AufloesungX;
end;

function TRakCanvasData.PixelToMmY(Pixel: integer;
  UseOffset: boolean): extended;
begin
  if UseOffset then
    Result := Pixel / AufloesungY - MmOffsetY
  else
    Result := Pixel / AufloesungY;
end;

procedure TRakCanvasData.SetAufloesung(const Value: extended);
begin
  AufloesungX := Value;
  AufloesungY := Value;
end;

procedure TRakCanvasData.SetAufloesungX(const Value: extended);
begin
  if Value <> 0 then
    FAufloesungX := Value;
end;

procedure TRakCanvasData.SetAufloesungY(const Value: extended);
begin
  if Value <> 0 then
    FAufloesungY := Value;
end;

procedure TRakCanvasData.SetMmOffsetX(const Value: extended);
begin
  FMmOffsetX := Value;
end;

procedure TRakCanvasData.SetMmOffsetY(const Value: extended);
begin
  FMmOffsetY := Value;
end;


end.
