unit RakSmallTable;

{///////////////////////////////////////////////////////////////////////////////
// Beschreibung : Die Unit RakSmallTable stellt Klassen zur Verfgung, mit denen
                  man Daten (Strings) in tabellarischer Form speichern kann.
// erstellt am  : ???
////////////////////////////////////////////////////////////////////////////////
// nderungen:
///////////////////////////////////////////////////////////////////////////////}
// Delphi 5 Implementation
//
// Copyright (c) 2000 by Richard Kaspar
// ("http://www.kasparsoft.de")
//
// Definitions:
// - "Package" refers to the collection of files distributed by
//   the Copyright Holder, and derivatives of that collection of
//   files created through textual modification.
// - "Standard Version" refers to such a Package if it has not
//   been modified, or has been modified in accordance with the
//   wishes of the Copyright Holder.
// - "Copyright Holder" is whoever is name in the copyright or
//   copyrights for the package.
// - "You" is you, if you're thinking about copying or distributing
//   this Package.
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Package"), to deal in the Package without restriction,
// including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the
// Package, and to permit persons to whom the Package is furnished
// to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Package.
//
// You may modify your copy of this Package in any way, provided
// that you insert a prominent notice in each changed file stating
// how and when you changed a file, and provided that you do at
// least one of the following:
//
// a) allow the Copyright Holder to include your modifications in
// the Standard Version of the Package.
//
// b) use the modified Package only within your corporation or
// organization.
//
// c) rename any non standard executables, units, and classes so
// the names do not conflict with standard executables, units, and
// classes, and provide a separate manual page that clearly documents
// how it differs from the standard version.
//
// d) make other distribution arrangements with the Copyright Holder.
//
// The name of the Copyright Holder may not be used to endorse or
// promote products derived from this Package without specific prior
// written permission.
//
// THE PACKAGE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// PACKAGE OR THE USE OR OTHER DEALINGS IN THE PACKAGE.
{$I RAK.INC}

interface

uses Windows, Classes, Graphics, SysUtils,
  MiscFunc,
  RakCollection;

type
  TRakTableCell = class;
  TRakTableRows = class;
  TRakTableDataCollumns = class;


{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakSmallTable
// Beschreibung : TRakSmallTable speichert Daten in tabellarischer Form.
                  Jede Zelle kann standardmig einen String speichern
                  Abgeleitete Klassen knnen auch andere Daten in einer
                  Zelle speichern, wenn sie fr die Verwaltung der Reihen
                  eine von TRakTableRows abgeleitete Klasse verwenden.
// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakSmallTable = class(TRakCollectionItem)
  private
    FTablename: string;
    procedure SetTablename(const Value: string);
  protected
    FCols: TRakTableDataCollumns;
    FRows: TRakTableRows;

    FOrderCol: string;
    FOrderType: string;

    procedure CompareItems(Sender: TObject; Item1, Item2: TRakCollectionItem;
      var CompResult: integer); virtual;

    function  GetColCount: integer;
    procedure SetColCount(Value: integer);
    function  GetRowCount: integer;
    procedure SetRowCount(Value: integer);
    function GetCells(Col, Row: integer): TRakTableCell;
    procedure SetCells(Col, Row: integer; Value: TRakTableCell);
  public
    constructor Create; override;
    destructor Destroy; override;

    // legt eine komplette Kopie der Tabelle an
    procedure Assign(Source: TRakCollectionItem); override;
    // Sucht eine Reihe anhand des Wertes einer bestimmten Spalte. Die Spalte sollte ein eindeutiger Schlssel sein. Rckgabe -1, wenn nicht gefunden
    function FindRow(ColName, Value: string): integer;
    // Gibt den Wert einer Zelle als String zurck
    function GetValue(RowId: integer; ColName: string): string;
    // Gibt eine Tabelle zurck, die eine Teilmenge dieser Tabelle ist. Die Reihen knnen nach einer Spalte sortiert werden
    function SelectAll(WhereField: string; IsValue: string;
      OrderByField: string): TRakSmallTable;

    // Die Anzahl der Spalten
    property ColCount: integer read GetColCount write SetColCount;
    // Die einzelnen Spalten
    property Cols: TRakTableDataCollumns read FCols;
    // Die Anzahl der Zeilen
    property RowCount: integer read GetRowCount write SetRowCount;
    // Die einzelnen Zeilen
    property Rows: TRakTableRows read FRows;
    // Alle Zellen der Tabelle
    property Cells[Col, Row: integer]: TRakTableCell read GetCells write SetCells; default;
    // Der frei whlbare Name der Tabelle
    property Tablename: string read FTablename write SetTablename;
  end;


{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakTableCell
// Beschreibung : TRakTableCell ist die Basisklasse einer Tabellenzelle
                  TRakTableCell speichert einen String
// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakTableCellClass = class of TRakTableCell;
  TRakTableCell = class(TRakCollectionItem)
  public
    Value: string;
    procedure Assign(Source: TRakCollectionItem); override;
  end;

{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakTableDataCollumn
// Beschreibung : TRakTableDataCollumn reprsentiert eine Spalte der Tabelle
                  Jede Spalte kann einen Namen haben. Es wird nicht
                  geprft, ob dieser eindeutig ist
// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakTableDataCollumnClass = class of TRakTableDataCollumn;
  TRakTableDataCollumn = class(TRakCollectionItem)
  protected
    FColName: string;
    FTable: TRakSmallTable;

    function GetCells(Row: integer): TRakTableCell; virtual;
    procedure SetCollection(Value: TRakCollection); override;
    procedure SetColName(Value: string); virtual;
  public
    constructor Create; override;

    // Assign bernimmt nur den Namen der Spalte, aber nicht die Werte der einzelnen Zellen
    procedure Assign(Source: TRakCollectionItem); override;

    // Die einzelnen Zellen in der Spalte
    property Cells[Row: integer]: TRakTableCell read GetCells;
    // Der Name der Spalte
    property ColName: string read FColName write SetColName;
  end;

{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakTableDataCollumns
// Beschreibung : TRakTableDataCollumns verwaltet alle Spalten einer Tabelle
// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakTableDataCollumns = class(TRakCollection)
  private
    function GetItems(Index: integer): TRakTableDataCollumn;
    procedure SetItems(Index: Integer; Value: TRakTableDataCollumn);
  protected
    FTable: TRakSmallTable;

    function GetCollumnClass: TRakTableDataCollumnClass; virtual;
    function AddTyped(Typ: TRakTableDataCollumnClass):
      TRakTableDataCollumn; virtual;
  public
    //Fgt eine neue Spalte hinzu und gibt ihr den Namen "ColName"
    function Add(ColName: string): TRakTableDataCollumn; virtual;
    //Sucht die Spalte mit einem bestimmten Namen; Rckage nil, falls nicht gefunden
    function ItemByName(ColName: string): TRakTableDataCollumn;

    // Alle Spalten der Tabelle (0 <= Index < Count)
    property Items[Index: Integer]: TRakTableDataCollumn read GetItems
      write SetItems; default;
end;


{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakTableDataRow
// Beschreibung : TRakTableDataRow reprsentiert eine Zeile einer Tabelle
// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakTableDataRowClass = class of TRakTableDataRow;
  TRakTableDataRow = class(TRakCollection)
  private
    function GetCells(Index: Integer): TRakTableCell;
    procedure SetCells(Index: Integer; Value: TRakTableCell);
  protected
    // Die Tabelle, zu der die Reihe gehrt
    FTable: TRakSmallTable;

    function AddEmptyCell: TRakTableCell; virtual;
    function GetCellClass: TRakTableCellClass; virtual;
    procedure SetCollection(Value: TRakCollection); override;
  public
    constructor Create; override;

    // Hngt hinten eine neue Zelle an; Es wird nicht geprft, ob die anderen Reihen der Tabelle genausoviele Zellen haben, wie diese Reihe
    function Add: TRakTableCell; virtual;

    // Die einzelnen Zellen in der Reihe (0 <= Index < Count)
    property Cells[Index: Integer]: TRakTableCell read GetCells
      write SetCells; default;
  end;


{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakTableRows
// Beschreibung : TRakTableRows verwaltet alle Reihen einer Tabelle
// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakTableRows = class(TRakCollection)
  private
    function GetItems(Index: Integer): TRakTableDataRow;
    procedure SetItems(Index: Integer; Value: TRakTableDataRow);
  protected
    FTable: TRakSmallTable;

    function GetRowClass: TRakTableDataRowClass; virtual;
    function AddTyped(Typ: TRakTableDataRowClass):
      TRakTableDataRow; virtual;
  public
    //Hngt hinten eine neue Reihe an; Die Reihe enthlt dann noch keine Zellen
    function Add: TRakTableDataRow; virtual;

    // Alle Reihen; (0 <= Index < Count)
    property Items[Index: Integer]: TRakTableDataRow read GetItems
      write SetItems; default;
  end;



implementation



constructor TRakSmallTable.Create;
begin
  inherited Create;

  FRows := TRakTableRows.Create;
  FRows.FTable := self;

  FCols := TRakTableDataCollumns.Create;
  FCols.FTable := self;
end;

destructor TRakSmallTable.Destroy;
begin
  FRows.Free;
  FCols.Free;

  inherited Destroy;
end;


procedure TRakSmallTable.Assign(Source: TRakCollectionItem);
begin
  inherited;

  FCols.Assign(TRakSmallTable(Source).FCols);
  FRows.Assign(TRakSmallTable(Source).FRows);
end;



function TRakSmallTable.GetCells(Col, Row: integer): TRakTableCell;
begin
  Result := FRows[Row][Col];
end;

procedure TRakSmallTable.SetCells(Col, Row: integer; Value: TRakTableCell);
begin
  FRows[Row][Col].Assign(Value);
end;


function  TRakSmallTable.GetColCount: integer;
begin
  Result := FCols.Count;
end;

procedure TRakSmallTable.SetColCount(Value: integer);
var
  i,t: integer;
begin
  for i := 0 to FRows.Count - 1 do
  begin
    if Value < FRows[i].Count then
    begin
      for t := FRows[i].Count -1 downto Value do
        FRows[i][t].Free;
    end
    else
    begin
      for t := 1 to Value - FRows[i].Count do
        FRows[i].AddEmptyCell;
    end;
  end;

  if Value < FCols.Count then
  begin
    for t := FCols.Count -1 downto Value do
      FCols[t].Free;
  end
  else
  begin
    for t := 1 to Value - FCols.Count do
      FCols.Add('');
  end;
end;

function  TRakSmallTable.GetRowCount: integer;
begin
  Result := FRows.Count;
end;

procedure TRakSmallTable.SetRowCount(Value: integer);
var
  i: integer;
begin
  if Value < FRows.Count then
  begin
    for i := FRows.Count -1 downto Value do
      FRows[i].Free;
  end
  else
  begin
    for i := 1 to Value - FRows.Count do
      FRows.Add;
  end;
end;







procedure TRakTableCell.Assign(Source: TRakCollectionItem);
begin
  inherited Assign(Source);

  Value := TRakTableCell(Source).Value;
end;



procedure TRakTableDataCollumn.Assign(Source: TRakCollectionItem);
begin
  inherited;

  FColName := TRakTableDataCollumn(Source).ColName;
end;

constructor TRakTableDataCollumn.Create;
begin
  inherited Create;

  FTable := nil;
end;

function TRakTableDataCollumn.GetCells(Row: integer): TRakTableCell;
begin
  if Assigned(FTable) then
    Result := FTable.GetCells(Index, Row)
  else
    Result := nil;
end;

procedure TRakTableDataCollumn.SetCollection(Value: TRakCollection);
begin
  if (Collection <> Value) then
  begin
    IgnoreChange := true;
    try
      inherited SetCollection(Value);

      if (Collection <> nil) and (Collection is TRakTableDataCollumns) then
      begin
        FTable := (Collection as TRakTableDataCollumns).FTable;
      end;
    finally
      IgnoreChange := false;
    end;
    Changed(self);
  end;
end;

procedure TRakTableDataCollumn.SetColName(Value: string);
begin
  FColName := Value;
end;


function TRakTableDataCollumns.GetItems(Index: integer): TRakTableDataCollumn;
begin
  Result := TRakTableDataCollumn(inherited GetItems(Index));
end;

procedure TRakTableDataCollumns.SetItems(Index: Integer;
  Value: TRakTableDataCollumn);
begin
  inherited SetItems(Index, Value);
end;


function TRakTableDataCollumns.Add(ColName: string): TRakTableDataCollumn;
begin
  Result := TRakTableDataCollumn(AddTyped(GetCollumnClass));
  Result.ColName := ColName;
end;

function TRakTableDataCollumns.ItemByName(ColName: string): TRakTableDataCollumn;
var
  i: integer;
begin
  Result := nil;
  i := 0;
  while (Result = nil) and (i < Count) do
  begin
    if CompareText(Items[i].ColName, ColName) = 0 then
      Result := Items[i];
    inc(i);
  end;
end;


constructor TRakTableDataRow.Create;
begin
  inherited Create;

  FTable := nil;
end;

function TRakTableDataRow.Add: TRakTableCell;
begin
  Result := TRakTableCell(BasicAdd(TRakTableCell));
end;


procedure TRakTableDataRow.SetCollection(Value: TRakCollection);
begin
  if (Collection <> Value) then
  begin
    IgnoreChange := true;
    try
      inherited SetCollection(Value);

      if (Collection <> nil) and (Collection is TRakTableRows) then
      begin
        FTable := (Collection as TRakTableRows).FTable;
      end;
    finally
      IgnoreChange := false;
    end;
    Changed(self);
  end;
end;


function TRakTableDataRow.GetCells(Index: Integer): TRakTableCell;
begin
  Result := TRakTableCell(inherited GetItems(Index));
end;

procedure TRakTableDataRow.SetCells(Index: Integer; Value: TRakTableCell);
begin
  inherited SetItems(Index, Value);
end;


function TRakTableRows.GetItems(Index: Integer): TRakTableDataRow;
begin
  Result := TRakTableDataRow(inherited GetItems(Index));
end;

procedure TRakTableRows.SetItems(Index: Integer; Value: TRakTableDataRow);
begin
  inherited SetItems(Index, Value);
end;

function TRakTableRows.Add: TRakTableDataRow;
begin
  Result := AddTyped(GetRowClass);
end;



procedure TRakSmallTable.SetTablename(const Value: string);
begin
  FTablename := Value;
end;

function TRakSmallTable.FindRow(ColName, Value: string): integer;
var
  i: integer;
  ColIdx: integer;
begin
  Result := -1;
  ColIdx := Cols.ItemByName(ColName).Index;
  if ColIdx >= 0 then
  begin
    i := 0;
    while (Result = -1) and (i < RowCount) do
    begin
      if Cells[ColIdx, i].Value = Value then
        Result := i;
      inc(i);
    end;
  end;
end;

function TRakSmallTable.GetValue(RowId: integer; ColName: string): string;
var
  ColId: integer;
begin
  Result := '';
  ColId := Cols.ItemByName(ColName).Index;
  if ColId > -1 then
    Result := Cells[ColId, RowId].Value;
end;

function TRakSmallTable.SelectAll(WhereField, IsValue,
  OrderByField: string): TRakSmallTable;
var
  i: integer;
  ColIdx: integer;
  Col: string;
  Wo: integer;
  ColName: string;
  ColType: string;
begin
  Result := TRakSmallTable.Create;
  try
    // Die Ergebnistabelle hat alle Spalten von dieser Tabelle
    Result.Cols.Assign(Cols);

    // Jetzt suchen wir nach den Datensaetzen
    ColIdx := Cols.ItemByName(WhereField).Index;
    if ColIdx >= 0 then // Wenn es die FilterSpalte gibt
    begin
      for i := 0 to RowCount - 1 do
      begin // Alle Reihen abklappern
        if Cells[ColIdx, i].Value = IsValue then
        begin // Wenn das ein Treffer ist
          Result.Rows.Add.Assign(Rows[i]);
        end;
      end;
    end;

    // Jetzt haben wir alle gesuchten Datensaetze
    // Die werde jetzt noch sortiert
    FOrderCol := '';
    FOrderType := '';
    for i := 0 to ListLen(OrderByField) - 1 do
    begin
      Col := ListGetAt(OrderByField, i);
      Wo := Pos(':', Col);
      if Wo = 0 then
      begin
        ColName := Col;
        ColType := 'string';
      end
      else
      begin
        ColName := ListGetAt(Col, 0, ':');
        ColType := ListGetAt(Col, 1, ':');
      end;
      if ColName <> '' then
      begin
        ColIdx := Cols.ItemByName(ColName).Index;
        if ColIdx > -1 then
        begin // Wenn es das Feld, nach dem sortiert werden soll gibt
          FOrderCol := ListAppend(FOrderCol, IntToStr(ColIdx));
          FOrderType := ListAppend(FOrderType, ColType);
        end;
      end;
    end;
    if FOrderCol <> '' then
    begin // Wenn es mindestens eine der Spalten gibt, nach denen
          // sortiert werden soll
      Result.Rows.OnCompareItems := CompareItems;
      Result.Rows.Sort;
    end;

  except
    Result.Free;
    raise;
  end;
end;

procedure TRakSmallTable.CompareItems(Sender: TObject; Item1,
  Item2: TRakCollectionItem; var CompResult: integer);
var
  I1, I2: TRakTableDataRow;
  i: integer;
  Idx: integer;
  ALen: integer;
begin
  I1 := Item1 as TRakTableDataRow;
  I2 := Item2 as TRakTableDataRow;
  CompResult := 0;
  i := 0;
  ALen := ListLen(FOrderCol);
  while (CompResult = 0) and (i < ALen) do
  begin
    Idx := StrToIntDef(ListGetAt(FOrderCol, i), -1);
    if Idx > -1 then
    begin
      if ListGetAt(FOrderType, i) = 'integer' then
      begin
        CompResult := StrToIntDef(I1[Idx].Value, Low(integer))
          - StrToIntDef(I2[Idx].Value, Low(integer));
      end
      else
        CompResult := CompareText(I1[Idx].Value, I2[Idx].Value);
    end;
    inc(i);
  end;
end;

function TRakTableDataCollumns.AddTyped(
  Typ: TRakTableDataCollumnClass): TRakTableDataCollumn;
begin
  Result := TRakTableDataCollumn(inherited Add(Typ));
  Result.FTable := FTable;
end;

function TRakTableDataCollumns.GetCollumnClass: TRakTableDataCollumnClass;
begin
  Result := TRakTableDataCollumn;
end;

function TRakTableRows.AddTyped(
  Typ: TRakTableDataRowClass): TRakTableDataRow;
var
  i: integer;
begin
  Result := TRakTableDataRow(BasicAdd(Typ));
  Result.FTable := FTable;
  for i := 1 to FTable.ColCount do
    Result.Add;
end;

function TRakTableRows.GetRowClass: TRakTableDataRowClass;
begin
  Result := TRakTableDataRow;
end;

function TRakTableDataRow.AddEmptyCell: TRakTableCell;
begin
  Result := Add;
end;

function TRakTableDataRow.GetCellClass: TRakTableCellClass;
begin
  Result := TRakTableCell;
end;

end.
