unit RakTableData;

{///////////////////////////////////////////////////////////////////////////////
// Beschreibung : Die Unit RakTableData stellt Klassen zur Verfgung, mit denen
                  man Daten in tabellarischer Form speichern kann.
                  Wenn man nur Strings ohne viel Komfort in Tabellenform
                  speichern mchte, dann sollte man besser die Klasse
                  TRakSmallTable aus der Unit RakSmallTable verwenden,
                  da diese viel kleiner und handlicher ist.
                  Die Klassen fressen ausserdem Unmengen an Resourcen. Sie sollten
                  also nur fr relativ kleine Tabellen verwendet werden
// erstellt am  : ???
////////////////////////////////////////////////////////////////////////////////
// nderungen:
///////////////////////////////////////////////////////////////////////////////}
// Delphi 5 Implementation
//
// Copyright (c) 2000 by Richard Kaspar
// ("http://www.kasparsoft.de")
//
// Definitions:
// - "Package" refers to the collection of files distributed by
//   the Copyright Holder, and derivatives of that collection of
//   files created through textual modification.
// - "Standard Version" refers to such a Package if it has not
//   been modified, or has been modified in accordance with the
//   wishes of the Copyright Holder.
// - "Copyright Holder" is whoever is name in the copyright or
//   copyrights for the package.
// - "You" is you, if you're thinking about copying or distributing
//   this Package.
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Package"), to deal in the Package without restriction,
// including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the
// Package, and to permit persons to whom the Package is furnished
// to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Package.
//
// You may modify your copy of this Package in any way, provided
// that you insert a prominent notice in each changed file stating
// how and when you changed a file, and provided that you do at
// least one of the following:
//
// a) allow the Copyright Holder to include your modifications in
// the Standard Version of the Package.
//
// b) use the modified Package only within your corporation or
// organization.
//
// c) rename any non standard executables, units, and classes so
// the names do not conflict with standard executables, units, and
// classes, and provide a separate manual page that clearly documents
// how it differs from the standard version.
//
// d) make other distribution arrangements with the Copyright Holder.
//
// The name of the Copyright Holder may not be used to endorse or
// promote products derived from this Package without specific prior
// written permission.
//
// THE PACKAGE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// PACKAGE OR THE USE OR OTHER DEALINGS IN THE PACKAGE.
{$I RAK.INC}

interface

uses Windows, Classes, Graphics, SysUtils,
{$IFDEF USE_WYSIWYGFONT}
  WYSIWYGFont,
  MiscFunc,
{$ENDIF}
  RakCollection;

type
  TFuncRec = class;
  TRakTableCell = class;
  TRakTableCellClass = class of TRakTableCell;
  TRakTableRows = class;
  TRakTableDataRow = class;
  TRakTableDataCollumn = class;
  TRakTableDataCollumns = class;
  TRakTableVersionException = class(Exception);

  THorzAusricht = (haLinks, haMittig, haRechts, haKomma, haLinksOffset,
    haRechtsOffset);
//  TNeededSpace = (nsMoreRows, nsMoreWidth, nsMoreHeight);
//  TNeededSpaces = set of TNeededSpace;


  // Die folgenden Aufzhlungstypen beeinflussen sich gegenseitig
  // Sie legen fest wie die Hhen und Breiten der Zellen einer
  // Tabelle an Grenvernderung der Tabelle angepat werden
  TAutoColumnWidth = (
    acwFix,  // Die Breiten der Spalten werden nicht verndert
    acwProportional, // die Breiten der Spalten werden im gleichen
                     // Verhltnis verndert wie sich die Breite
                     // der Tabelle ndert
    acwAuto, // Die Breiten der Spalten werden so gro gemacht, da sie
             // mglichst viel des zur Verfgung stehenden Platzes ausfllen
             // und das Verhltnis der Breiten, dem der Breiten von
             // den jeweils breitesten Werten enspricht
    acwMax,  // wie acwAuto, auer da die Breiten der Spalten die maximale
             // Breite einer Spalte nicht berschreiten darf
    acwValue // Die Breiten der Spalten werden so gro gemacht, da ihre
             // Inhalte gerade dargestellt werden knnen, ohne da rechts oder
             // links etwas abgeschnitten wird
    );
  TAutoRowHeight = (
    arhFix,  // Die Hhe der Reihen wird nicht verndert
    arhMal,  // Die Hhe der Reihen mu ein Vielfaches von DefaultRowHeight
             // sein. Das Vielfache wird von der Anzahl der Umbrche
             // bestimmt (siehe TAutoValueWidth)
    arhAuto, // Die Hhe der Reihen wird so gro gemacht, da sie
             // mglichst viel des zur Verfgung stehenden Platzes ausfllen
    arhMax,  // wie arhAuto, auer da die Hhe der Reihen die maximale
             // Hhe einer Reihe (MaxRowHeight) nicht berschreiten darf
    arhValue // Die Hhe der Reihen wird so gro gemacht, da ihre Inhalte
             // gerade dargestellt werden knnen, ohne da unten oder
             // oben etwas abgeschnitten wird
             // arhValue verhlt sich wie arhFix, wenn nicht 'avhFix'
             // oder 'avhWidth' aktiv ist
    );
  TAutoValueHeight = (
    avhFix,  // Die Hhe der Werte wird nicht verndert
             // Wenn der Wert zu Hoch ist, dann wird er abgeschnitten
    avhAuto, // Die Hhe der Werte wird so gro gemacht, da sie
             // mglichst so gro sind, wie die Reihe hoch ist
    avhMax,  // wie avhAuto, auer da die Hhe der Werte die maximal
             // mgliche Hhe nicht berschreiten darf
    avhWidth // Die Hhe der Werte richtet sich nach deren Breite
    );
  TAutoValueWidth = (
    avwFix,  // Die Breite der Werte wird nicht verndert
             // Wenn der Wert zu Breit ist, dann wird er abgeschnitten
{$IFDEF USE_WYSIWYGFONT}
    avwAuto, // Die Breite der Werte wird so gro gemacht, da sie
             // mglichst so gro sind, wie die Spalte breit ist
    avwMax,  // wie avwAuto, auer da die Breite der Werte die maximal
             // mgliche Breite nicht berschreiten darf
    avwStauchen,// wie avwMax, wobei die maximale Breite sich
                // aus dem natrlichen Verhltnis zur Hhe ergibt
{$ENDIF}
    avwUmbruch  // Wenn der Wert zu breit ist, dann wird er 'zerschnitten'
                // und die einzelnen Teile untereinander angeordnet.
                // Das drfte wohl nur fr Text sinnvoll sein
{$IFDEF USE_WYSIWYGFONT}
    ,
    avwStauchenUmbruch// Eine Mischung aus avwStauchen und avwUmbruch.
                       // Die Breite der Werte wird solange verringert
                       // bis sie in die Spalte passen, oder ein gewisses
                       // Verhltnis zur Hhe unterschritten wird. Dann
                       // wird umgebrochen
{$ENDIF}
    );
    // Die folgenden Werte sind nur relevant, wenn 'avwUmbruch'
    // oder 'avwStauchenUmbruch' aktiv ist
  TAutoUmbruch = (
    auFix,  // Wenn der Wert umgebrochen wird, dann bleibt die Hhe der
            // Reihe unverndert. Die Hhe des Wertes wird soweit verringert,
            // da er komplett dargestellt werden kann
            // Die Hhe des Wertes wird nur dann verringert, wenn nicht
            // 'avhFix' eingestellt ist
    auBig   // Die Hhe der betroffenen Reihe wird um einen Fakor vergrert.
            // Der Faktor ist die Anzahl der erforderlichen Umbrche
    );

  // Die Zusammenhnge der oben aufgefhrten Aufzhlungstypen wird mir jetzt
  // zu kompliziert, deswegen wird nun ein einzelner Aufzhlungstyp
  // fr die Formatierung eingefhrt. Jedes einzelne Element stellt dann
  // eine Kombination der oben aufgefhrten Aufzhlungstypen dar. Die
  // Formatierung jedes einzelnen Elements wird separat durchgefhrt, damit
  // der Alghorithmus nicht unendlich komplex wird.
  TAutoFormatType = (
    aftNone,          // Darstellung als Tabelle.
                      // Es werden die momentan aktiven Einstellungen genommen
    aftTableMaxWidth, // Darstellung als Tabelle.
                      // Die Werte werden so breit wie mglich gemacht.
                      // Das natrliche Verhltnis Breite zu Hhe wird beibehalten
    aftTableMaxHeightStauch  // Darstellung als Tabelle.
                      // Die Werte werden so hoch wie mglich gemacht.
                      // Sollte Der Text dadurch zu breit sein, so wird er in der
                      // Breite gestaucht
    );


  TCellChangedEvent = procedure(Sender: TObject; Cell: TRakTableCell) of object;

{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakCustomTableData ist die Basisklasse aller Tabellen,
                  die ihre Daten im Speicher halten.
                  TRakCustomTableData selbst speichert keine Daten. Erst davon
                  abgeleitete Klassen sind in der Lage wirklich Daten zu
                  speichern (siehe Klasse TRakStringTable).
                  TRakCustomTableData kann sich selbst zeichnen und in einem
                  Stream speichern. Weiterhin kann es den Inhalt seiner Zellen
                  formatieren.
// Beschreibung :

// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakCustomTableData = class
  private
    FFAutoFormatType: TAutoFormatType;
  protected
    FAutoRowHeight: TAutoRowHeight;
    FAutoColumnWidth: TAutoColumnWidth;
    FAutoValueHeight: TAutoValueHeight;
    FAutoValueWidth: TAutoValueWidth;
    FAutoUmbruch: TAutoUmbruch;
    FAutoFormatType: TAutoFormatType;
    FMaxColumnWidth: integer;

    FDrehung: integer;

    FSaveAutoRowHeight: TAutoRowHeight;
    FSaveAutoColumnWidth: TAutoColumnWidth;
    FSaveAutoValueHeight: TAutoValueHeight;
    FSaveAutoValueWidth: TAutoValueWidth;
    FSaveAutoUmbruch: TAutoUmbruch;

    FCols: TRakTableDataCollumns;
    FMinRowHeight: integer;
    FMaxRowHeight: integer;
    FRows: TRakTableRows;

    FOnCellChanged: TCellChangedEvent;

    procedure AutoColWidths(Canvas: TCanvas; Breite: integer); virtual;
    procedure AutoCellsValueHeihgt(Canvas: TCanvas); virtual;
    procedure AutoRowHeights(Canvas: TCanvas; Hoehe: integer); virtual;
    procedure AutoSize(Canvas: TCanvas; Hoehe, Breite: integer); virtual;

    procedure CreateFCols; virtual;
    procedure CreateFRows; virtual;
    function GetCellType: TRakTableCellClass; virtual; abstract;

    procedure NormalizeCellSize; virtual;

    function GetCellsObject(Col, Row: integer): TRakTableCell;
    procedure SetCellsObject(Col, Row: integer; Value: TRakTableCell);

//    function GetAutoRowHeight: TAutoRowHeight; virtual;
    procedure SetAutoRowHeight(Value: TAutoRowHeight); virtual;
//    function GetAutoColumnWidth: TAutoColumnWidth; virtual;
    procedure SetAutoColumnWidth(Value: TAutoColumnWidth); virtual;
//    function GetAutoValueHeight: TAutoValueHeight; virtual;
    procedure SetAutoValueHeight(Value: TAutoValueHeight); virtual;
//    function GetAutoValueWidth: TAutoValueWidth; virtual;
    procedure SetAutoValueWidth(Value: TAutoValueWidth); virtual;
//    function GetAutoUmbruch: TAutoUmbruch; virtual;
    procedure SetAutoUmbruch(Value: TAutoUmbruch); virtual;
    procedure SetAutoFormatType(const Value: TAutoFormatType); virtual;
    procedure SetDrehung(const Value: integer); virtual;

    function  GetColAusrichtung(Id: integer): THorzAusricht; virtual;
    procedure SetColAusrichtung(Id: integer; Value: THorzAusricht); virtual;
    function  GetColCount: integer; virtual;
    procedure SetColCount(Value: integer); virtual;
    function  GetColWidths(Id: integer): integer; virtual;
    procedure SetColWidths(Id: integer; Value: integer); virtual;
    function  GetRowCount: integer; virtual;
    procedure SetRowCount(Value: integer); virtual;

    property AutoRowHeight: TAutoRowHeight read FAutoRowHeight
      write SetAutoRowHeight;
    property AutoColumnWidth: TAutoColumnWidth read FAutoColumnWidth
      write SetAutoColumnWidth;
    property AutoValueHeight: TAutoValueHeight read FAutoValueHeight
      write SetAutoValueHeight;
    property AutoValueWidth: TAutoValueWidth read FAutoValueWidth
      write SetAutoValueWidth;
    property AutoUmbruch: TAutoUmbruch read FAutoUmbruch write SetAutoUmbruch;
    property AutoFormatType: TAutoFormatType read FFAutoFormatType
      write SetAutoFormatType;

    property MaxColumnWidth: integer read fMaxColumnWidth write FMaxColumnWidth;

    property Drehung: integer read FDrehung write SetDrehung;

    property CellsObject[Col, Row: integer]: TRakTableCell read GetCellsObject
      write SetCellsObject;
  public
    constructor Create;
    destructor Destroy; override;

    // Assign legt eine Kopie der Tabell an
    procedure Assign(Source: TRakCustomTableData); virtual;

    procedure SetAutoParams; virtual;
    procedure SaveAutoParams; virtual;
    procedure RestoreAutoParams; virtual;

    // Ldt die Daten der Tabelle aus einem Stream
    procedure LoadAsStream(Reader: TReader); virtual;
    // Schreibt die Daten der Tabelle in einen Stream
    procedure SaveAsStream(Writer: TWriter); virtual;

    // AutoFormat formatiert den Inhalt der Zellen anhand des fr die Ausgabe zur verfgungstehenden Platzes und der AutoFormat-Properties
    procedure AutoFormat(Canvas: TCanvas; Hoehe, Breite: integer); virtual;
    // AutoFormatByType formatiert den Inhalt der Zellen anhand des fr die Ausgabe zur verfgungstehenden Platzes und des Properties AutoFormatType
    procedure AutoFormatByType(Canvas: TCanvas; Hoehe, Breite: integer); virtual;

    // (protected) Changed lst das Ereignis OnCellChanged aus
    procedure Changed(Cell: TRakTableCell); virtual;

    // Paint zeichnet die Tabelle
    procedure Paint(Canvas: TCanvas; X, Y: integer); virtual;

    // Legt fr jede Spalte fest, wie sich die Werte der Zellen ausrichten sollen
    property ColAusrichtung[Id: integer]: THorzAusricht read GetColAusrichtung
      write SetColAusrichtung;
    // Die Anzahl der Spalten
    property ColCount: integer read GetColCount write SetColCount;
    // Alle Spalten der Tabelle
    property Cols: TRakTableDataCollumns read FCols;
    // Die Breite jeder einzelnen Spalte
    property ColWidths[Id: integer]: integer read GetColWidths
      write SetColWidths;
    // Die Mindesthhe einer Reihe
    property MinRowHeight: integer read FMinRowHeight write FMinRowHeight;
    // Die maximale Hhe einer Reihe
    property MaxRowHeight: integer read FMaxRowHeight write FMaxRowHeight;
    // die Anzahl der Reihen
    property RowCount: integer read GetRowCount write SetRowCount;
    // Alle Reihen der Tabelle
    property Rows: TRakTableRows read FRows;

    // Ereignis, das ausgelst wird, wenn sich der Wert einer Zelle ndert
    property OnCellChanged: TCellChangedEvent read FOnCellChanged
      write FOnCellChanged;
  end;


  TRakTableCellLine = class;
  TCellLastChange = (clcData, clcLine, clcBrush, clcPen, clcFont,
    clcAusrichtung, clcOffset, clcElse, clcElse0, clcElse1, clcElse2,
    clcElse3, clcElse4, clcElse5, clcElse6, clcElse7, clcElse8, clcElse9);
    // Die 'clcElseX (X = 0 - 9)' knnen ersetzt
    // werden indem sie eine Konstante dafr bereitstellen
    // z.B. :  const clcNeuerGrund = clcElse0;

{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakTableCell ist die Basisklasse aller Zellen einer Tabelle
                  TRakTableCell selbst speichert keine Daten. Das mssen davon
                  abgeleitete Klassen machen, siehe Klasse TRakStringCell.
// Beschreibung :

// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakTableCell = class(TRakCollectionItem)
  private
    MaxHeight: integer;
{$IFDEF USE_WYSIWYGFONT}
    MaxWidth: integer;
{$ENDIF}
    AutoValueHeight: TAutoValueHeight;
    AutoValueWidth: TAutoValueWidth;
    AutoUmbruch: TAutoUmbruch;
    ZeilenCount: integer;
    NeededHeight: integer;
    NeededWidth: integer;
  protected
    FBrush: TBrush;
    FHorzAusricht: THorzAusricht;
    FHorzOffset: integer;
    FLastChange: TCellLastChange;
    FLineLeft: TRakTableCellLine;
    FLineTop: TRakTableCellLine;
    FLineRight: TRakTableCellLine;
    FLineBottom: TRakTableCellLine;

    function AutoFormat(Canvas: TCanvas; MaxHeightFor_avhWidth: integer): boolean; virtual;
    function GetValueHeight(Canvas: TCanvas): integer; virtual;

    procedure GraphicObjectChanged(Sender: TObject); virtual;
    procedure LineChanged(Line: TRakTableCellLine); virtual;

    function GetAsString: string; virtual; abstract;
    procedure SetAsString(Value: string); virtual; abstract;
    procedure SetHorzAusricht(Value: THorzAusricht); virtual;
    procedure SetHorzOffset(Value: integer); virtual;
    function GetParentCol: TRakTableDataCollumn; virtual;
    function GetParentRow: TRakTableDataRow; virtual;
    function GetParentTable: TRakCustomTableData; virtual;
  public
    constructor Create; override;
    destructor Destroy; override;

    procedure Assign(Source: TRakCollectionItem); override;
    // Gibt den Breite zurck, welche die Zelle zur Ausgabe des Nachkommawertes bentigt, falls die Zelle eine Zahl speichert
    function GetValueNachKommaWidth(Canvas: TCanvas): integer; virtual;
    // Gibt die Breite zurck, welche die Zelle zur Ausgabe ihres Wertes bentigt
    function GetValueWidth(Canvas: TCanvas): integer; virtual;
    // Gibt den Breite zurck, welche die Zelle zur Ausgabe des Vorkommawertes bentigt, falls die Zelle eine Zahl speichert
    function GetValueVorKommaWidth(Canvas: TCanvas): integer; virtual;
    // Zeichnet die Zelle, Inhalt und Rahmen ohne Hintergrund
    procedure Paint(Canvas: TCanvas; ClientRect: TRect;
      Drehung: integer); virtual;
    // Zeichnet den Hintergrund der Zelle
    procedure PaintBackground(Canvas: TCanvas; ClientRect: TRect;
      Drehung: integer); virtual;

    //Ldt die Daten der Zelle aus einem Stram
    procedure LoadAsStream(Reader: TReader); virtual;
    //Speichert die Daten der Zelle in einem Stram
    procedure SaveAsStream(Writer: TWriter); virtual;

    //Zugriff des Wertes in Stringform
    property AsString: string read GetAsString write SetAsString;

    // Der Pinsel fr die Ausgabe des Hintergunds
    property Brush: TBrush read FBrush;
    // die  horizontale Ausrichtung des Wertes
    property HorzAusricht: THorzAusricht read FHorzAusricht
      write SetHorzAusricht;
    // wirkt sich zusammen mit HorzAusricht auf die Ausgabe des Wertes aus
    property HorzOffset: integer read FHorzOffset write SetHorzOffset;
    // Gibt den Grund an, warum das Ereignis TRakCustomTableData.OnCellChange ausgelst wurde
    property LastChange: TCellLastChange read FLastChange;
    // Linie des linken Randes der Zelle
    property LineLeft: TRakTableCellLine read FLineLeft;
    // Linie des oberen Randes der Zelle
    property LineTop: TRakTableCellLine read FLineTop;
    // Linie des rechten Randes der Zelle
    property LineRight: TRakTableCellLine read FLineRight;
    // Linie des unteren Randes der Zelle
    property LineBottom: TRakTableCellLine read FLineBottom;
    // Die Spalte, in der sich die Zelle befindet
    property ParentCol: TRakTableDataCollumn read GetParentCol;
    // Die Reihe, in der sich die Zelle befindet
    property ParentRow: TRakTableDataRow read GetParentRow;
    // Die Tabelle, zu der die Zelle gehrt
    property ParentTable: TRakCustomTableData read GetParentTable;
  end;

  TRakTableCellLineTyp = (ltNone, ltSingle);

{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakTableCellLine wird fr den Rand einer Zelle eingesetzt
// Beschreibung :

// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakTableCellLine = class(TPersistent)
  protected
    FPen: TPen;
    FParentCell: TRakTableCell;
    FTyp: TRakTableCellLineTyp;
    FVisible: boolean;

    procedure PenChange(Sender: TObject); virtual;
    procedure SetTyp(Value: TRakTableCellLineTyp); virtual;
    procedure SetVisible(Value: boolean); virtual;
  public
    constructor Create(Parent: TRakTableCell);
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;
    procedure Changed; virtual;
    procedure Paint(Canvas: TCanvas; ClientRect: TRect;
      Seite: integer); virtual;

    procedure LoadAsStream(Reader: TReader); virtual;
    procedure SaveAsStream(Writer: TWriter); virtual;

    // Der Stift, mit dem sich die Linie Zeichnet
    property Pen: TPen read FPen;
    // Die Zelle, zu der die Linie gehrt
    property ParentCell: TRakTableCell read FParentCell;
    // Typ legt fest, welche Linie des Zellenrandes es ist
    property Typ: TRakTableCellLineTyp read FTyp write SetTyp;
    // Legt fest, ob sich die Linie zeichnet
    property Visible: boolean read FVisible write SetVisible;
  end;

{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakTableDataCollumn reprsentiert eine Spalte einer Tabelle
// Beschreibung :

// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakTableDataCollumnClass = class of TRakTableDataCollumn;
  TRakTableDataCollumn = class(TRakCollectionItem)
  private
    MaxWidth: integer;
  protected
    FAusrichtung: THorzAusricht;
    FBreiteAbschneidbar: boolean;
    FColName: string;
    FLine: TRakTableCellLineTyp;
    FTable: TRakCustomTableData;
    FVisible: boolean;
    FWidth: integer;

    procedure SetAusrichtung(Value: THorzAusricht); virtual;
    function GetCells(Row: integer): TRakTableCell; virtual;
    procedure SetCollection(Value: TRakCollection); override;
    procedure SetColName(Value: string); virtual;
    procedure SetLine(Value: TRakTableCellLineTyp); virtual;
    function GetVisible: boolean; virtual;
    procedure SetVisible(Value: boolean); virtual;
    function GetVisibleIndex: integer; virtual;
    procedure SetWidth(Value: integer); virtual;
  public
    constructor Create; override;

    procedure Assign(Source: TRakCollectionItem); override;
    // Setzt die Breite der Spalte so, dass sie proportional zu den Breiten der anderen Spalten ist
    procedure AutoWidth(Canvas: TCanvas); virtual;

    procedure LoadAsStream(Reader: TReader); virtual;
    procedure SaveAsStream(Writer: TWriter); virtual;

    // Legt fest, wie sich die Inhalte der Zellen ausrichten
    property Ausrichtung: THorzAusricht read FAusrichtung write SetAusrichtung;
    // Legt fest, ob Zellen abgeschnitten werden drfen, falls ihre Werte breiter sind, als die zur Verfgung stehende Breite
    property BreiteAbschneidbar: boolean read FBreiteAbschneidbar
      write FBreiteAbschneidbar;
    //Die Zellen der Spalte
    property Cells[Row: integer]: TRakTableCell read GetCells;
    // Der Name der Spalten
    property ColName: string read FColName write SetColName;
    // Die Linie, mit der sich die Spalte von der nchsten Spalte abtrennt
    property Line: TRakTableCellLineTyp read FLine write SetLine;
    // Gibt an, ob sich die Spalte gezeichnzet werden darf
    property Visible: boolean read GetVisible write SetVisible;
    // Gibt an, die wievielte sichtbare Spalte dies ist
    property VisibleIndex: integer read GetVisibleIndex;
    // Die Breite der Spalte
    property Width: integer read FWidth write SetWidth;
  end;

{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakTableDataCollumns verwaltet alle Spalten einer Tabelle
// Beschreibung :

// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakTableDataCollumns = class(TRakCollection)
  private
    function GetItems(Index: integer): TRakTableDataCollumn;
    procedure SetItems(Index: Integer; Value: TRakTableDataCollumn);
    function GetVisibleItems(Idx: Integer): TRakTableDataCollumn;
  protected
    FTable: TRakCustomTableData;
  public
    // Eine neue Spalte hinzufgen
    function Add: TRakTableDataCollumn;
    // Eine neue Spalte von einem ganz bestimmten Typ hinzufgen
    function AddTyped(Typ: TRakTableDataCollumnClass):
      TRakTableDataCollumn; virtual;
    // Anzahl aller sichtbaren Spalten
    function CountVisible: integer; virtual;
    // Gibt an, die wievielte sichtbare Spalte dies ist
    function GetVisibleIndex(Item: TRakTableDataCollumn): integer; virtual;
    //Sucht nach einer Spalte anhand ihres Namens
    function ItemByName(ColName: string): TRakTableDataCollumn;

    procedure LoadAsStream(Reader: TReader); virtual;
    procedure SaveAsStream(Writer: TWriter); virtual;

    // Alle Spalten (0 <= Index < Count)
    property Items[Index: Integer]: TRakTableDataCollumn read GetItems
      write SetItems; default;
    // Alle Sichtbaren Spalten (0 <= Idx < CountVisible)
    property VisibleItems[Idx: Integer]: TRakTableDataCollumn
      read GetVisibleItems;
  end;


{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakTableDataRow reprsentiert eine Reihe einer Tabelle
// Beschreibung :

// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakTableDataRow = class(TRakCollection)
  private
    DefaultHeight: integer;
    MaxHeight: integer;

    function GetCells(Index: Integer): TRakTableCell;
    procedure SetCells(Index: Integer; Value: TRakTableCell);
  protected
    FHeight: integer;
    FLine: TRakTableCellLineTyp;
    FTable: TRakCustomTableData;

    procedure SetCollection(Value: TRakCollection); override;
    procedure SetHeight(Value: integer); virtual;
    procedure SetLine(Value: TRakTableCellLineTyp); virtual;
  public
    constructor Create; override;

    // Fgt der Reihe eine neue Zelle hinzu
    function Add(TheClass: TRakCollectionItemClass): TRakCollectionItem;
    procedure Assign(Source: TRakCollectionItem); override;

    procedure LoadAsStream(Reader: TReader); virtual;
    procedure SaveAsStream(Writer: TWriter); virtual;

    // Die Hhe der Reihe
    property Height: integer read FHeight write SetHeight;
    // Alle Zellen der Reihe
    property Cells[Index: Integer]: TRakTableCell read GetCells
      write SetCells; default;
    // Die Linie, mit der sich die Reihe von der nchsten abtrennt
    property Line: TRakTableCellLineTyp read FLine write SetLine;
  end;


{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakTableRows verwaltet alle Reihen einer Tabelle
// Beschreibung :

// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakTableRows = class(TRakCollection)
  private
    function GetItems(Index: Integer): TRakTableDataRow;
    procedure SetItems(Index: Integer; Value: TRakTableDataRow);
  protected
    FTable: TRakCustomTableData;
  public
    // Fgt eine neue Reihe hinzu
    function Add: TRakTableDataRow; virtual;

    procedure LoadAsStream(Reader: TReader); virtual;
    procedure SaveAsStream(Writer: TWriter); virtual;

    // Alle Reihen der Tabelle
    property Items[Index: Integer]: TRakTableDataRow read GetItems
      write SetItems; default;
  end;

{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakStringCell
// Beschreibung : TRakStringCell reprsentiert eine Zelle einer Tabelle vom
                  Typ TRakStringTable. Die Zelle speichert einen String.
                  Zur Ausgabe des Strings kann ein normaler Font (TFont)
                  oder eine Wysiwyg-Font (TRakWYSIWYGFont aus Unit WYSIWYGFont)
                  verwendet werden
// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakStringCell = class(TRakTableCell)
  private
{$IFDEF USE_WYSIWYGFONT}
    FDoClip: boolean;
    procedure SetDoClip(const Value: boolean);
{$ENDIF}
  protected
    FCanvas: TCanvas;
    FFont: TFont;
    FFuncRec: TFuncRec;
    FOnlyWholeChars: boolean;
    FValue: string;

{$IFDEF USE_WYSIWYGFONT}
    FUseWYSIWYGFont: boolean;
    FWYSIWYGFont: TRakWYSIWYGFont;
{$ENDIF}

    function AutoFormat(Canvas: TCanvas; MaxHeightFor_avhWidth: integer): boolean; override;

    function TextHeight(Zeile: string): integer; virtual;
    function TextWidth(Zeile: string): integer; virtual;

    function GetAsString: string; override;
    procedure SetAsString(Value: string); override;
    function GetAsInteger: integer; virtual;
    procedure SetAsInteger(Value: integer); virtual;
    procedure SetValue(Value: string); virtual;

    function  GetFontSize: integer; virtual;
    procedure SetFontSize(Value: integer); virtual;
{$IFDEF USE_WYSIWYGFONT}
    function  GetFontWidth: integer;
    procedure SetFontWidth(Value: integer);
{$ENDIF}


    property FontSize: integer read GetFontSize write SetFontSize;
{$IFDEF USE_WYSIWYGFONT}
    property FontWidth: integer read GetFontWidth write SetFontWidth;
{$ENDIF}
  public
    constructor Create; override;
    destructor Destroy; override;

    procedure Assign(Source: TRakCollectionItem); override;
    function GetValueNachKommaWidth(Canvas: TCanvas): integer; override;
    function GetValueWidth(Canvas: TCanvas): integer; override;
    function GetValueHeight(Canvas: TCanvas): integer; override;
    function GetValueVorKommaWidth(Canvas: TCanvas): integer; override;
    procedure Paint(Canvas: TCanvas; ClientRect: TRect;
      Drehung: integer); override;

    procedure LoadAsStream(Reader: TReader); override;
    procedure SaveAsStream(Writer: TWriter); override;

    // Der Font, mit der String ausgegeben wird
    property Font: TFont read FFont;
    // Legt fest, ob Zeichen ausgegeben werden, die ber den Zellenrand hinausgehen
    property OnlyWholeChars: boolean read FOnlyWholeChars
      write FOnlyWholeChars;

    // Der Wert als Zahl
    property AsInteger: integer read GetAsInteger write SetAsInteger;
    property Value: string read FValue write SetValue;

{$IFDEF USE_WYSIWYGFONT}
    property UseWYSIWYGFont: boolean read FUseWYSIWYGFont write FUseWYSIWYGFont;
    property WYSIWYGFont: TRakWYSIWYGFont read FWYSIWYGFont;
    property DoClip: boolean read FDoClip write SetDoClip;
{$ENDIF}
  end;

{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakStringTable
// Beschreibung : TRakStringTable speichert Strings in tabellarischer Form
// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakStringTable = class(TRakCustomTableData)
  protected
    FCanvas: TCanvas;
    FFont: TFont;
    FFuncRec: TFuncRec;
    FOnlyWholeChars: boolean;

{$IFDEF USE_WYSIWYGFONT}
    FUseWYSIWYGFont: boolean;
    FWYSIWYGFont: TRakWYSIWYGFont;
{$ENDIF}
    FSaveFontWidth: integer;


    function GetCellType: TRakTableCellClass; override;

    procedure FontChanged(Sender: TObject); virtual;

    procedure NormalizeCellSize; override;

    function TextHeight(Zeile: string): integer; virtual;
    function TextWidth(Zeile: string): integer; virtual;

    function GetAsInteger(Col, Row: integer): integer; virtual;
    procedure SetAsInteger(Col, Row: integer; Value: integer); virtual;

    function GetCellsObject(Col, Row: integer): TRakStringCell;
    function GetCells(Col, Row: integer): string;
    procedure SetCells(Col, Row: integer; Value: string);
    function GetVisibleCells(Col, Row: integer): string;
    procedure SetVisibleCells(Col, Row: integer; Value: string);

    procedure SetColCount(Value: integer); override;
    procedure SetOnlyWholeChars(Value: boolean); virtual;
    procedure SetRowCount(Value: integer); override;

    function  GetFontSize: integer; virtual;
    procedure SetFontSize(Value: integer); virtual;
{$IFDEF USE_WYSIWYGFONT}
    function  GetFontWidth: integer;
    procedure SetFontWidth(Value: integer);
{$ENDIF}

    property FontSize: integer read GetFontSize write SetFontSize;
  public
    constructor Create;
    destructor Destroy; override;

    procedure Assign(Source: TRakCustomTableData); override;

    procedure AutoCellsValueHeihgt(Canvas: TCanvas); override;
    procedure AutoSize(Canvas: TCanvas; Hoehe, Breite: integer); override;

    procedure LoadAsStream(Reader: TReader); override;
    procedure SaveAsStream(Writer: TWriter); override;

    procedure SetAutoParams; override;
    procedure SaveAutoParams; override;
    procedure RestoreAutoParams; override;

    property AutoRowHeight;
    property AutoColumnWidth;
    property AutoValueHeight;
    property AutoValueWidth;
    property AutoUmbruch;

    property AutoFormatType;
    property Drehung;

    property AsInteger[Col, Row: integer]: integer read GetAsInteger
      write SetAsInteger;

    property CellsObject[Col, Row: integer]: TRakStringCell read GetCellsObject;
    property Cells[Col, Row: integer]: string read GetCells
      write SetCells; default;
    property VisibleCells[Col, Row: integer]: string read GetVisibleCells
      write SetVisibleCells;

    property Font: TFont read FFont;
    property OnlyWholeChars: boolean read FOnlyWholeChars
      write SetOnlyWholeChars;

{$IFDEF USE_WYSIWYGFONT}
    property UseWYSIWYGFont: boolean read FUseWYSIWYGFont write FUseWYSIWYGFont;
    property WYSIWYGFont: TRakWYSIWYGFont read FWYSIWYGFont;
{$ENDIF}
  end;


  TGetTextHeight = function(Zeile: string): integer of object;
  TGetTextWidth = function(Zeile: string): integer of object;
  TGetFontSize = function: integer of object;
  TSetFontSize = procedure(Value: integer) of object;

  TFuncRec = class
  private
    FGetTextHeight: TGetTextHeight;
    FGetTextWidth: TGetTextWidth;
    FGetFontSize: TGetFontSize;
    FSetFontSize: TSetFontSize;
{$IFDEF USE_WYSIWYGFONT}
    FGetFontWidth: TGetFontSize;
    FSetFontWidth: TSetFontSize;
{$ENDIF}

    function lGetFontSize: integer;
    procedure lSetFontSize(Value: integer);
{$IFDEF USE_WYSIWYGFONT}
    function lGetFontWidth: integer;
    procedure lSetFontWidth(Value: integer);
{$ENDIF}
  public
    function TextHeight(Zeile: string): integer;
    function TextWidth(Zeile: string): integer;

    property FontSize: integer read lGetFontSize write lSetFontSize;
{$IFDEF USE_WYSIWYGFONT}
    property FontWidth: integer read lGetFontWidth write lSetFontWidth;
{$ENDIF}

    property GetTextHeight: TGetTextHeight read FGetTextHeight
      write FGetTextHeight;
    property GetTextWidth: TGetTextWidth read FGetTextWidth
      write FGetTextWidth;
    property GetFontSize: TGetFontSize read FGetFontSize write FGetFontSize;
    property SetFontSize: TSetFontSize read FSetFontSize write FSetFontSize;
{$IFDEF USE_WYSIWYGFONT}
    property GetFontWidth: TGetFontSize read FGetFontWidth write FGetFontWidth;
    property SetFontWidth: TSetFontSize read FSetFontWidth write FSetFontWidth;
{$ENDIF}
  end;


procedure FitFontSize(FuncRec: TFuncRec; MaxHeight: integer);
procedure FitFontSizeByWidth(FuncRec: TFuncRec; MaxWidth: integer; Zeile: string);

{$IFDEF USE_WYSIWYGFONT}
procedure FitFontWidth(FuncRec: TFuncRec; MaxWidth: integer; Zeile: string);
{$ENDIF}

implementation



constructor TRakCustomTableData.Create;
begin
  inherited Create;

  CreateFRows;
  FRows.FTable := self;
  CreateFCols;
  FCols.FTable := self;
  FMaxRowHeight := high(integer);
  FMinRowHeight := 1;
  FFAutoFormatType := aftNone;
end;

destructor TRakCustomTableData.Destroy;
begin
  FRows.Free;
  FCols.Free;

  inherited Destroy;
end;

procedure TRakCustomTableData.CreateFCols;
begin
  FCols := TRakTableDataCollumns.Create;
end;

procedure TRakCustomTableData.CreateFRows;
begin
  FRows := TRakTableRows.Create;
end;


procedure TRakCustomTableData.Assign(Source: TRakCustomTableData);
begin
  AutoRowHeight := Source.AutoRowHeight;
  AutoColumnWidth := Source.AutoColumnWidth;
  AutoValueHeight := Source.AutoValueHeight;
  AutoValueWidth := Source.AutoValueWidth;
  AutoUmbruch := Source.AutoUmbruch;
  MaxColumnWidth := Source.MaxColumnWidth;
  FCols.Assign(Source.FCols);
  FMaxRowHeight := Source.FMaxRowHeight;
  FMinRowHeight := Source.FMinRowHeight;
  FRows.Assign(Source.FRows);
end;

procedure TRakCustomTableData.LoadAsStream(Reader: TReader);
var
  iWert: integer;
begin
  iWert := Reader.ReadInteger;
  if iWert > 2 then
    raise TRakTableVersionException.Create('Version 2');

  FAutoRowHeight := TAutoRowHeight(Reader.ReadInteger);
  FAutoColumnWidth := TAutoColumnWidth(Reader.ReadInteger);
  FAutoValueHeight := TAutoValueHeight(Reader.ReadInteger);
  FAutoValueWidth := TAutoValueWidth(Reader.ReadInteger);
  FAutoUmbruch := TAutoUmbruch(Reader.ReadInteger);
  FMaxColumnWidth := Reader.ReadInteger;
  FMaxRowHeight := Reader.ReadInteger;
  if iWert > 1 then
    FMinRowHeight := Reader.ReadInteger;

  FCols.LoadAsStream(Reader);
  FRows.LoadAsStream(Reader);
end;

procedure TRakCustomTableData.SaveAsStream(Writer: TWriter);
begin
  Writer.WriteInteger(2);

  Writer.WriteInteger(Ord(FAutoRowHeight));
  Writer.WriteInteger(Ord(FAutoColumnWidth));
  Writer.WriteInteger(Ord(FAutoValueHeight));
  Writer.WriteInteger(Ord(FAutoValueWidth));
  Writer.WriteInteger(Ord(FAutoUmbruch));
  Writer.WriteInteger(FMaxColumnWidth);
  Writer.WriteInteger(FMaxRowHeight);
  Writer.WriteInteger(FMinRowHeight);
  FCols.SaveAsStream(Writer);
  FRows.SaveAsStream(Writer);
end;

procedure TRakCustomTableData.AutoFormat(Canvas: TCanvas;
  Hoehe, Breite: integer);

  procedure Format;
    procedure CellFormat;
    var
      AktRow: TRakTableDataRow;
//      AktCol: TRakTableDataCollumn;
      AktCell: TRakTableCell;
      Row, Col: integer;
//      NewVal: integer;
      Changed: boolean;
      MaxHeightFor_avhWidth: integer;
      Col_avhWidth: TRakIntegerCollection;
    begin // CellFormat
      Col_avhWidth := TRakIntegerCollection.Create;
      try
        if (AutoValueHeight = avhWidth) and (AutoColumnWidth = acwValue) then
        begin
          Col_avhWidth.Count := ColCount;
          if RowCount > 0 then
          begin
            for Col := 0 to ColCount - 1 do
              Col_avhWidth[Col] := Rows[0].Height;
          end;
        end;
        // Die folgende Schleife wird solange durchlaufen, bis die Zellinhalte
        // sich  mglichst gut an die gegebenen Hhen und Breiten angepat haben
        repeat
          //Merken, da sich bis jetzt noch nichts gendert hat
          Changed := false;
          MaxHeightFor_avhWidth := 0;

          Row := 0;
          while Row < RowCount do
          begin // Fr jede Reihe
            // Die aktuelle Reihe merken
            AktRow := Rows[Row];
            Col := 0;
            MaxHeightFor_avhWidth := AktRow.Height;
            while Col < ColCount do
            begin // Fr jede Zelle der aktuellen Reihe
              // Die aktuelle Spalte merken
  //            AktCol := Cols[Col];
              // Die aktuelle Zelle merken
              AktCell := CellsObject[Col, Row];
              // Die Zelle darf sich selbst formatieren
              AktCell.AutoFormat(Canvas, MaxHeightFor_avhWidth);

              if AutoValueHeight = avhWidth then
              begin
                if AutoRowHeight = arhValue then
                begin
                  if (MaxHeightFor_avhWidth > AktCell.NeededHeight)
                    and (AktCell.NeededHeight > 0) then
                  begin
                    MaxHeightFor_avhWidth := AktCell.NeededHeight;
                  end;
                end;

                if AutoColumnWidth = acwValue then
                begin
                  if AktCell.NeededHeight > 0 then
                  begin
                    if Col_avhWidth[Col] > AktCell.NeededHeight then
                      Col_avhWidth[Col] := AktCell.NeededHeight;
                    if MaxHeightFor_avhWidth > AktCell.NeededHeight then
                      MaxHeightFor_avhWidth := AktCell.NeededHeight;
                  end;
                end;
              end;
  {
              // Die Hhe der Reihe an die Hhe der Zelle anpassen
              if AktCell.NeededHeight > AktRow.Height then
              begin
                case AutoRowHeight of
                arhMal: // Wenn die Hhe der Reihen sich nur stufenweiese verndern darf
                  begin
                    if AktCell.NeededHeight mod AktRow.DefaultHeight > 0 then
                    begin // Wenn die bentigte Hhe exakt eine der erlaubten Hhen ist
                      if AktRow.Height <> AktCell.NeededHeight then
                      begin // Wenn sich die Hhe der Reihe wirklich verndern wrde
                        AktRow.Height := AktCell.NeededHeight;
                        Changed := true;
                      end;
                    end
                    else
                    begin // Wenn die bentigte Hhe nicht exakt eine der erlaubten Hhen ist
                      // Die nchste grere erlaubte Hhe berechenen
                      NewVal := (AktCell.NeededHeight div AktRow.DefaultHeight
                        + 1) * AktRow.DefaultHeight;
                      if AktRow.Height <> NewVal then
                      begin // Wenn sich die Hhe der Reihe wirklich verndern wrde
                        AktRow.Height := NewVal;
                        Changed := true;
                      end;
                    end;
                  end;
                arhAuto, arhValue: // Wenn die Hhe der Reihen sich stufenlos und beliebig gro verndern darf
                  begin
                    AktRow.Height := AktCell.NeededHeight;
                    Changed := true;
                  end;
                arhMax: // Wenn die Hhe der Reihen sich stufenlos aber nicht beliebig gro verndern darf
                  begin
                    if AktCell.NeededHeight > AktRow.MaxHeight then
                    begin // Wenn die bentigte Hhe grer als erlaubt ist
                      if AktRow.Height <> AktRow.MaxHeight then
                      begin // Wenn sich die Hhe der Reihe wirklich verndern wrde
                        AktRow.Height := AktRow.MaxHeight;
                        Changed := true;
                      end;
                    end
                    else
                    begin
                      AktRow.Height := AktCell.NeededHeight;
                      Changed := true;
                    end;
                  end;
                end; // case
              end;

              // Die Breite der Spalte an die Breite der Zelle anpassen
              if AktCell.NeededWidth > AktCol.Width then
              begin
                case AutoColumnWidth of
                acwAuto, acwValue:
                  begin
                    AktCol.Width := AktCell.NeededWidth;
                    Changed := true;
                  end;
                acwMax:
                  begin
                    if AktCell.NeededWidth > AktCol.MaxWidth then
                    begin
                      if AktCol.Width <> AktCol.MaxWidth then
                      begin
                        AktCol.Width := AktCol.MaxWidth;
                        Changed := true;
                      end;
                    end
                    else
                    begin
                      AktCol.Width := AktCell.NeededWidth;
                      Changed := true;
                    end;
                  end;
                end; // case
              end;

              if Changed and (Col = 0) and (Row = 0) then
                Changed := false;
              }
              inc(Col);
            end;

            if AutoValueHeight = avhWidth then
            begin // Wenn sich die Hhe der Schrift nach der Breite richten soll
              for Col := 0 to ColCount - 1 do
              begin
                // Die aktuelle Zelle merken
                AktCell := CellsObject[Col, Row];

                if MaxHeightFor_avhWidth < AktCell.NeededHeight then
                  AktCell.AutoFormat(Canvas, MaxHeightFor_avhWidth);
              end;
            end;

            inc(Row);
          end;

          if (AutoValueHeight = avhWidth) and (AutoColumnWidth = acwValue) then
          begin // Wenn sich die Hhe der Schrift nach der Breite richten soll
                // und die Spaltenbreiten nach ihren Werten
            if AutoRowHeight = arhValue then
            begin // Wenn sich auch die Reihenhhen nach ihren Werten richten sollen
              if ColCount > 0 then
              begin
                MaxHeightFor_avhWidth := Col_avhWidth[0];
                for Col := 1 to ColCOunt -1 do
                begin
                  if MaxHeightFor_avhWidth > Col_avhWidth[Col] then
                    MaxHeightFor_avhWidth := Col_avhWidth[Col];
                end;
              end;
              for Col := 0 to ColCount - 1 do
              begin
                for Row := 0 to RowCount - 1 do
                begin
                  // Die aktuelle Zelle merken
                  AktCell := CellsObject[Col, Row];

                  if MaxHeightFor_avhWidth < AktCell.NeededHeight then
                    AktCell.AutoFormat(Canvas, MaxHeightFor_avhWidth);
                end;
              end;
            end
            else
            begin  // Wenn sich die Reihenhhen nicht nach ihren Werten richten sollen
              for Col := 0 to ColCount - 1 do
              begin
                for Row := 0 to RowCount - 1 do
                begin
                  // Die aktuelle Zelle merken
                  AktCell := CellsObject[Col, Row];

                  if Col_avhWidth[Col] < AktCell.NeededHeight then
                    AktCell.AutoFormat(Canvas, Col_avhWidth[Col]);
                end;
              end;
            end;
          end;
        until not Changed;
      finally
        Col_avhWidth.Free;
      end;
    end; // CellFormat

    procedure RowFormat;
      procedure FormatArhMal;
      var
        Row, Col: integer;
        MaxCount: integer;
      begin //FormatArhMal
        for Row := 0 to RowCount - 1 do
        begin
          MaxCount := 1;

          // Feststellen, wieviele Zeilenumbrche fr die Reihe gebrucht werden
          for Col := 0 to ColCount - 1 do
          begin
            if CellsObject[Col, Row].ZeilenCount > MaxCount then
              MaxCount := CellsObject[Col, Row].ZeilenCount;
          end;

          if Rows[Row].Height <> MaxCount * Rows[Row].DefaultHeight then
          begin
            Rows[Row].Height := MaxCount * Rows[Row].DefaultHeight;
          end;
        end;
      end; //FormatArhMal

      procedure FormatVar;
        function ShrinkHeight: integer;
        var
          MaxHeight: integer;
          Row, Col: integer;
        begin //ShrinkHeight
          Result := 0;

          for Row := 0 to RowCount - 1 do
          begin
            MaxHeight := 1;

            // Ermitteln wieviel Platz die hchste Zelle in der aktuellen Reihe bentigt
            for Col := 0 to ColCount - 1 do
            begin
              if CellsObject[Col, Row].NeededHeight > MaxHeight then
                MaxHeight := CellsObject[Col, Row].NeededHeight;
            end;

            if Rows[Row].Height <> MaxHeight then
            begin // Falls die aktuelle Reihe eine Hhe hat, die nicht genau
                  // ausreicht um den grten Wert anzuzeigen
              if (AutoRowHeight <> arhMax)
                or (Rows[Row].MaxHeight >= MaxHeight) then
              begin // Wenn die Reihe nicht auf eine maximale Hhe begrenzt ist
                    // und die maximale Hhe nicht berschritten wird, um alles
                    // anzeigen zu knnen
                Rows[Row].Height := MaxHeight;
              end
              else if Rows[Row].MaxHeight > Rows[Row].Height then
              begin // Wenn die maximale Hhe berschritten wrde, aber
                    // die Reihe noch niedriger als die maximale Hhe ist
                Rows[Row].Height := Rows[Row].MaxHeight;
              end;
            end;

            Result := Result + Rows[Row].Height;
          end;
        end; //ShrinkHeight

        procedure Scale(AllHeight: integer);
        var
          OldHeight, NewHeight: integer;
          Row: integer;
        begin //Scale
          if AllHeight <> 0 then
          begin
            OldHeight := 0;
            for Row := 0 to RowCount - 1 do
            begin
              NewHeight := trunc(OldHeight / AllHeight * Hoehe);
              OldHeight := OldHeight + Rows[Row].Height;
              NewHeight := trunc(OldHeight / AllHeight * Hoehe) - NewHeight;
              if NewHeight <> Rows[Row].Height then
              begin
                Rows[Row].Height := NewHeight;
//                Fertig := false;
              end;
            end;
          end;
        end; //Scale

        procedure SizeMaxHeight;
        var
          AllHeight, OldHeight, NewHeight: integer;
          Row: integer;
          Weiter: boolean;
        begin //SizeMaxHeight
          if AutoRowHeight = arhMax then
          begin
            Weiter := true;
            while Weiter do
            begin
              Weiter := false;
              AllHeight := 0;
              for Row := 0 to RowCount - 1 do
              begin
                if Rows[Row].Height <> Rows[Row].MaxHeight then
                  AllHeight := AllHeight + Rows[Row].Height;
              end;

              if AllHeight <> 0 then
              begin
                OldHeight := 0;
                for Row := 0 to RowCount - 1 do
                begin
                  if Rows[Row].Height <> Rows[Row].MaxHeight then
                  begin
                    NewHeight := trunc(OldHeight / AllHeight * Hoehe);
                    OldHeight := OldHeight + Rows[Row].Height;
                    NewHeight := trunc(OldHeight / AllHeight * Hoehe)
                      - NewHeight;

                    if NewHeight < Rows[Row].MaxHeight then
                      Rows[Row].Height := NewHeight
                    else
                      Rows[Row].Height := Rows[Row].MaxHeight;
                  end
                  else
                  begin
                    Weiter := true;
                  end;
                end;
              end;
            end;
          end;
        end; //SizeMaxHeight

      var
        AllHeight: integer;
      begin // FormatVar
        // Alle Reihen an die Hhe der Werte anpassen und ermitteln,
        // wie hoch die Tabelle dann wrde
        AllHeight := ShrinkHeight;

        if AutoRowHeight <> arhValue then
        begin // Wenn
          if AllHeight <> Hoehe then
          begin // Wenn die Tabellenhhe nicht der Hhe aller Reihen entspricht
            if AutoRowHeight = arhAuto then
            begin // Wenn sich die Hhen der Reihen automatisch an die Hhe
                  // der Tabelle anpassen sollen
              Scale(AllHeight);
            end
            else
            begin // Wenn sich die Hhen der Reihen automatisch an die Hhe
                  // der Tabelle anpassen sollen, wobei eine maximale Reihenhhe
                  // nicht berschritten werden darf
              if AllHeight > Hoehe then
              begin // Wenn die Reihenhhen verkleinert werden mssen
                Scale(AllHeight)
              end
              else // Wenn die Reihenhhen vergrert werden mssen
                SizeMaxHeight;
            end;
          end;
        end
        else if AutoValueHeight = avhWidth then
        begin
        end;
      end; // FormatVar

    begin // RowFormat
      case AutoRowHeight of
      arhMal: FormatArhMal;
      arhAuto, arhValue, arhMax: FormatVar;
      end;
    end; // RowFormat

    procedure ColFormat;
      procedure FormatVar;
        function ShrinkWidth: integer;
        var
          MaxWidth: integer;
          Row, Col: integer;
        begin //ShrinkWidth
          Result := 0;

          for Col := 0 to ColCount - 1 do
          begin
            MaxWidth := 1;

            // Ermitteln wieviel Platz die breiteste Zelle in der aktuellen Spalte bentigt
            for Row := 0 to RowCount - 1 do
            begin
              if CellsObject[Col, Row].NeededWidth > MaxWidth then
                MaxWidth := CellsObject[Col, Row].NeededWidth;
            end;

            if Cols[Col].Width <> MaxWidth then
            begin // Falls die aktuelle Reihe eine Hhe hat, die nicht genau
                  // ausreicht um den grten Wert anzuzeigen
              if (AutoColumnWidth <> acwMax)
                or (Cols[Col].MaxWidth >= MaxWidth) then
              begin // Wenn die Reihe nicht auf eine maximale Hhe begrenzt ist
                    // und die maximale Hhe nicht berschritten wird, um alles
                    // anzeigen zu knnen
                Cols[Col].Width := MaxWidth;
              end
              else if Cols[Col].MaxWidth > Cols[Col].Width then
              begin // Wenn die maximale Hhe berschritten wrde, aber
                    // die Reihe noch niedriger als die maximale Hhe ist
                Cols[Col].Width := Cols[Col].MaxWidth;
              end;
            end;

            Result := Result + Cols[Col].Width;
          end;
        end; //ShrinkWidth

        procedure Scale(AllWidth: integer);
        var
          OldWidth, NewWidth: integer;
          Col: integer;
        begin //Scale
          if AllWidth <> 0 then
          begin
            OldWidth := 0;
            for Col := 0 to ColCount - 1 do
            begin
              NewWidth := trunc(OldWidth / AllWidth * Breite);
              OldWidth := OldWidth + Cols[Col].Width;
              NewWidth := trunc(OldWidth / AllWidth * Breite) - NewWidth;
              if NewWidth <> Cols[Col].Width then
              begin
                Cols[Col].Width := NewWidth;
//                Fertig := false;
              end;
            end;
          end;
        end; //Scale

        procedure SizeMaxWidth;
        var
          AllWidth, OldWidth, NewWidth: integer;
          Col: integer;
          Weiter: boolean;
        begin //SizeMaxWidth
          if AutoColumnWidth = acwMax then
          begin
            Weiter := true;
            while Weiter do
            begin
              Weiter := false;
              AllWidth := 0;
              for Col := 0 to ColCount - 1 do
              begin
                if Cols[Col].Width <> Cols[Col].MaxWidth then
                  AllWidth := AllWidth + Cols[Col].Width;
              end;

              if AllWidth <> 0 then
              begin
                OldWidth := 0;
                for Col := 0 to ColCount - 1 do
                begin
                  if Cols[Col].Width <> Cols[Col].MaxWidth then
                  begin
                    NewWidth := trunc(OldWidth / AllWidth * Breite);
                    OldWidth := OldWidth + Cols[Col].Width;
                    NewWidth := trunc(OldWidth / AllWidth * Breite)
                      - NewWidth;

                    if NewWidth < Cols[Col].MaxWidth then
                      Cols[Col].Width := NewWidth
                    else
                      Cols[Col].Width := Cols[Col].MaxWidth;
                  end
                  else
                  begin
                    Weiter := true;
                  end;
                end;
              end;
            end;
          end;
        end; //SizeMaxWidth

      var
        AllWidth: integer;
      begin // FormatVar
        // Alle Spalten an die Breite der Werte anpassen und ermitteln,
        // wie breit die Tabelle dann wrde
        AllWidth := ShrinkWidth;

        if AutoColumnWidth <> acwValue then
        begin
          if AllWidth <> Breite then
          begin // Wenn die Tabellenbreite nicht der Breite aller Spalten entspricht
            if AutoColumnWidth = acwAuto then
            begin // Wenn sich die Breiten der Spalten automatisch an die Breite
                  // der Tabelle anpassen sollen
              Scale(AllWidth);
            end
            else
            begin // Wenn sich die Breiten der Spalten automatisch an die Breite
                  // der Tabelle anpassen sollen, wobei eine maximale Spaltenbreite
                  // nicht berschritten werden darf
              if AllWidth > Breite then
              begin // Wenn die Spaltenbreiten verkleinert werden mssen
                Scale(AllWidth)
              end
              else // Wenn die Spaltenbreiten vergrert werden mssen
                SizeMaxWidth;
            end;
          end;
        end;
      end; // FormatVar

    begin //ColFormat
      case AutoColumnWidth of
      acwAuto, acwMax, acwValue: FormatVar;
      end;
    end; //ColFormat
  var
    Fertig: boolean;
    FOldRows, FOldCols: TRakIntegerCollection;
    i: integer;
  begin // Format
    FOldRows := nil;
    FOldCols := nil;
    try
      FOldRows := TRakIntegerCollection.Create;
      FOldCols := TRakIntegerCollection.Create;

      // Die folgende Schleife wird so oft durchlaufen, bis die Formatierung
      // pat. Da bei entsprechenden Optionen, die Abhngigkeiten der Hhen,
      // Breiten und Zeilenumbrchen so aufwendig sein kann, da sie durch
      // keine 'Formel' berechnet werden kann, mu man sich durch ein numerisches
      // Verfahren an den Idealfall herantasten
      repeat
        // In FOldRows werden sich die Hhen der Reihen gemerkt, um nach der
        // einem Formatierungsversuch zu berprfen, ob sich berhaupt noch etwas
        // verndert hat
        FOldRows.FreeAll;
        for i := 0 to RowCount - 1 do
          FOldRows.Add(Rows[i].Height);

        // In FOldCols werden sich die Breiten der Spalten gemerkt.
        FOldCols.FreeAll;
        for i := 0 to ColCount - 1 do
          FOldCols.Add(Cols[i].Width);

{$IFDEF USE_WYSIWYGFONT}
        if not(AutoValueWidth in [avwAuto, avwMax, avwStauchen]) then
{$ENDIF}
        begin
          // Die Zellinhalte ausgehend von den momentanen Breiten und Hhen
          // etwas nher an den Idealzustand bringen
          CellFormat;
        end;

        // Die Hhen der Reihen anhand der Zellinhalte formatieren
        RowFormat;

        // Die Breiten der Spalten anhand der Zellinhalte formatieren
        ColFormat;

{$IFDEF USE_WYSIWYGFONT}
        if AutoValueWidth in [avwAuto, avwMax, avwStauchen] then
        begin
          // Die Zellinhalte ausgehend von den momentanen Breiten und Hhen
          // etwas nher an den Idealzustand bringen
          CellFormat;
        end;
{$ENDIF}

        // Jetzt wird berprft, ob sich durch die letzte Formatierung
        // berhaupt noch etwas gendert hat. Wenn nicht, dann haben wir den
        // Idealzustand erreicht
        Fertig := true;
        if not (AutoRowHeight in [arhFix])
          and
          ( not (AutoRowHeight in [arhAuto, arhMax, arhValue])
            or
            (AutoValueWidth in [avwUmbruch
{$IFDEF USE_WYSIWYGFONT}
            , avwStauchenUmbruch
{$ENDIF}
            ])
          ) then
        begin
          for i := 0 to RowCount - 1 do
          begin
            if FOldRows[i] <> Rows[i].Height then
            begin
              Fertig := false;
              break;
            end;
          end;

          if Fertig then
          begin
            for i := 0 to ColCount - 1 do
            begin
              if FOldCols[i] <> Cols[i].Width then
              begin
                Fertig := false;
                break;
              end;
            end;
          end;
        end;

      until Fertig;
    finally;
      FOldRows.Free;
      FOldCols.Free;
    end;
  end; // Format
var
  Row, Col: integer;
  Wert: integer;
begin //TRakCustomTableData.AutoFormat
  if (ColCount > 0) and (RowCount > 0) then
  begin // Wenn es berhaupt etwas zum formatieren gibt

    // Zuerst wird die Breite der Spalten vorbereitet
    case AutoColumnWidth of
    acwFix, acwProportional:; // Der ideale Fall: an der Breite mu nichts
                              // verndert werden. Bei acwFix wurde die Breite
                              // durch den Anwender bestimmt. Bei acwProportional
                              // wird die Breite nur verndert, wenn sich
                              // die Gre der Tabelle ndert, und das passiert
                              // an einer anderen Stelle
    acwAuto, acwMax:  // Hier wird erst einmal die insgesamt zur Verfgung
                      // stehende Breite gleichmig verteilt. Spter
                      // wird die Breite dann noch an die des Inhalts angepat
      begin
        Wert := Breite div ColCount;
        if (AutoColumnWidth = acwMax) and (Wert > MaxColumnWidth) then
          Wert := MaxColumnWidth;
        for Col := 0 to ColCount - 1 do
          Cols[Col].Width := Wert;
      end;
    acwValue:   // Die Breite mu sich nach dem Inhalt richten. Da sich die
                // Inhalte aber wahrscheinlich nach der Spalte richten,
                // wird erst einmal jede Spalte gleich breit gemacht
      begin
        Wert := Breite div ColCount;
        for Col := 0 to ColCount - 1 do
          Cols[Col].Width := Wert;
      end;
    end;

    // Jetzt wird die Hhe der Spalten vorbereitet
    case AutoRowHeight of
    arhFix:;  // Der ideale Fall: an der Hhe mu nichts verndert werden.
    arhAuto, arhMax:  // Die Hhe der Reihen kann hier komplett festgelegt,
                      // da keine Anpassung en die Hhe der Inhalte erfolgen mu
      begin
        Wert := Hoehe div RowCount;
        if (AutoRowHeight = arhMax) and (Wert > MaxRowHeight) then
          Wert := MaxRowHeight;
        for Row := 0 to RowCount - 1 do
          Rows[Row].Height := Wert;
      end;
    arhValue:   // Die Hhe mu sich nach dem Inhalt richten. Da sich die
                // Inhalte aber wahrscheinlich nach der Reihe richten,
                // wird erst einmal jede Reihe gleich gro gemacht
      begin
        Wert := Hoehe div RowCount;
        for Row := 0 to RowCount - 1 do
          Rows[Row].Height := Wert;
      end;
    end;

    case AutoValueWidth of
    avwFix:;
{$IFDEF USE_WYSIWYGFONT}
    avwAuto, avwMax, avwStauchen:
      begin
        SaveAutoParams;

        if AutoColumnWidth in [acwAuto, acwMax, acwValue] then
        begin
          SetAutoParams;
          AutoValueWidth := avwFix;
          try
            Format;
          finally
            AutoValueWidth := FSaveAutoValueWidth;
            RestoreAutoParams;
          end;
        end;
      end;
{$ENDIF}
    avwUmbruch:;
{$IFDEF USE_WYSIWYGFONT}
    avwStauchenUmbruch:;
{$ENDIF}
    end;

    // Hhe und Breiten der Spalten und Reihen wurde vorbereitet
    // Darauf aufsetzend wird jetzt die weitere Formatierung durchgfhrt
    Format;
  end;
end; //TRakCustomTableData.AutoFormat


procedure TRakCustomTableData.AutoFormatByType(Canvas: TCanvas;
  Hoehe, Breite: integer);

  procedure Format_aftTableMaxWidth;
  var
    Row, Col: integer;
    ColWidths: TRakIntegerCollection;
    iWidth: integer;
    MaxWidth: integer;
    VWidth, RWidth: integer;
    MinHeight: integer;
  begin
    if RowCount > 0 then
    begin
      ColWidths := TRakIntegerCollection.Create;
      try
        ColWidths.Count := ColCount;
        NormalizeCellSize;
        // Fr jede Spalte feststellen, wie breit der breiteste Eintrag ist
        MaxWidth := 0;
        for Col := 0 to ColCount -1 do
        begin
          ColWidths[Col] := CellsObject[Col, 0].GetValueWidth(Canvas);
          for Row := 1 to RowCount -1 do
          begin
            iWidth := CellsObject[Col, Row].GetValueWidth(Canvas);
            if ColWidths[Col] < iWidth then
              ColWidths[Col] := iWidth;
          end;
          inc(MaxWidth, ColWidths[Col]);
        end;
        //Die Spaltenbreiten relativ zu ihrer gewnschten Breite und des zur
        // Verfgung stehenden Platzes einstellen
        VWidth := 0;
        RWidth := 0;
        for Col := 0 to ColCount - 1 do
        begin
          inc(VWidth, ColWidths[Col]);
          Cols[Col].Width := round(VWidth / MaxWidth * Breite) - RWidth;
          inc(RWidth, Cols[Col].Width);
        end;

        //Die Spaltenwerte an die Spaltenbreite anpassen
        //und die kleinste Hhe feststellen
        MinHeight := Hoehe div RowCount;
{$IFDEF USE_WYSIWYGFONT}
        AutoValueWidth := avwMax;
{$ENDIF}
        AutoValueHeight := avhWidth;
        for Row := 0 to RowCount -1 do
        begin
          for Col := 0 to ColCount - 1 do
          begin
            CellsObject[Col, Row].AutoFormat(Canvas, High(integer));
            if (CellsObject[Col, Row].NeededHeight > 0)
              and (MinHeight > CellsObject[Col, Row].NeededHeight) then
            begin
              MinHeight := CellsObject[Col, Row].NeededHeight;
            end;
          end;
        end;

        // Die Reihenhhe auf die kleinste Hhe einstellen
        for Row := 0 to RowCount -1 do
        begin
          Rows[Row].Height := MinHeight;
        end;

        // Und zum Schlu noch die Gren der werte richtig einstellen
        AutoValueHeight := avhAuto;
        AutoValueWidth := avwFix;
        for Row := 0 to RowCount -1 do
          for Col := 0 to ColCount - 1 do
            CellsObject[Col, Row].AutoFormat(Canvas, High(integer));
      finally
        ColWidths.Free;
      end;
    end;
  end;

  procedure Format_aftTableMaxHeightStauch;
  var
    Row, Col: integer;
    ColWidths: TRakIntegerCollection;
    iWidth: integer;
    MaxWidth: integer;
    VWidth, RWidth: integer;
    MinHeight: integer;
  begin
    if RowCount > 0 then
    begin
      ColWidths := TRakIntegerCollection.Create;
      try
        ColWidths.Count := ColCount;
        NormalizeCellSize;
        // Fr jede Spalte feststellen, wie breit der breiteste Eintrag ist
        MaxWidth := 0;
        for Col := 0 to ColCount -1 do
        begin
          ColWidths[Col] := CellsObject[Col, 0].GetValueWidth(Canvas);
          for Row := 1 to RowCount -1 do
          begin
            iWidth := CellsObject[Col, Row].GetValueWidth(Canvas);
            if ColWidths[Col] < iWidth then
              ColWidths[Col] := iWidth;
          end;
          inc(MaxWidth, ColWidths[Col]);
        end;
        //Die Spaltenbreiten relativ zu ihrer gewnschten Breite und des zur
        // Verfgung stehenden Platzes einstellen
        VWidth := 0;
        RWidth := 0;
        for Col := 0 to ColCount - 1 do
        begin
          inc(VWidth, ColWidths[Col]);
          Cols[Col].Width := round(VWidth / MaxWidth * Breite) - RWidth;
          inc(RWidth, Cols[Col].Width);
        end;

        //Die Spaltenwerte an die Spaltenbreite anpassen
        //und die kleinste Hhe feststellen
        MinHeight := Hoehe div RowCount;
        AutoValueWidth := avwFix;
{$IFDEF USE_WYSIWYGFONT}
        AutoValueWidth := avwStauchen;
{$ENDIF}
        AutoValueHeight := avhAuto;

        // Die Reihenhhe auf die kleinste Hhe einstellen
        for Row := 0 to RowCount -1 do
        begin
          Rows[Row].Height := MinHeight;
        end;

        // Und zum Schlu noch die Gren der werte richtig einstellen
        for Row := 0 to RowCount -1 do
          for Col := 0 to ColCount - 1 do
            CellsObject[Col, Row].AutoFormat(Canvas, 0);
      finally
        ColWidths.Free;
      end;
    end;
  end;
begin
  case  AutoFormatType of
    aftTableMaxWidth: Format_aftTableMaxWidth;
    aftTableMaxHeightStauch: Format_aftTableMaxHeightStauch;
  end;
end;


procedure TRakCustomTableData.SetAutoParams;
begin
end;

procedure TRakCustomTableData.SaveAutoParams;
begin
  FSaveAutoRowHeight := AutoRowHeight;
  FSaveAutoColumnWidth := AutoColumnWidth;
  FSaveAutoValueHeight := AutoValueHeight;
  FSaveAutoValueWidth := AutoValueWidth;
  FSaveAutoUmbruch := AutoUmbruch;
end;

procedure TRakCustomTableData.RestoreAutoParams;
begin
  AutoRowHeight := FSaveAutoRowHeight;
  AutoColumnWidth := FSaveAutoColumnWidth;
  AutoValueHeight := FSaveAutoValueHeight;
  AutoValueWidth := FSaveAutoValueWidth;
  AutoUmbruch := FSaveAutoUmbruch;
end;


procedure TRakCustomTableData.Changed(Cell: TRakTableCell);
begin
  if Assigned(FOnCellChanged) then
    FOnCellChanged(self, Cell);
end;

procedure TRakCustomTableData.Paint(Canvas: TCanvas; X, Y: integer);
var
  Col, Row: integer;
  DX, DY: integer;
  RX, RY: integer;
begin
  DY := Y;
  for Row := 0 to RowCount - 1 do
  begin
    DX := X;
    for Col := 0 to ColCount - 1 do
    begin
      RX := DX;
      RY := DY;
{$IFDEF USE_WYSIWYGFONT}
      Drehe(RX, RY, X, Y, Drehung);
{$ENDIF}
      CellsObject[Col,Row].PaintBackground(Canvas, Rect(RX, RY,
        RX + ColWidths[Col] - 1, RY + Rows[Row].Height - 1), Drehung);
      DX := DX + ColWidths[Col];
    end;
    DY := DY + Rows[Row].Height;
  end;

  DY := Y;
  for Row := 0 to RowCount - 1 do
  begin
    DX := X;
    for Col := 0 to ColCount - 1 do
    begin
      RX := DX;
      RY := DY;
{$IFDEF USE_WYSIWYGFONT}
      Drehe(RX, RY, X, Y, Drehung);
{$ENDIF}
      CellsObject[Col,Row].Paint(Canvas, Rect(RX, RY, RX + ColWidths[Col] - 1,
        RY + Rows[Row].Height - 1), Drehung);
      DX := DX + ColWidths[Col];
    end;
    DY := DY + Rows[Row].Height;
  end;
end;

procedure TRakCustomTableData.AutoColWidths(Canvas: TCanvas; Breite: integer);
var
  Col, NextSize: integer;
  Stauchbreite: integer;
  Stauchen: boolean;
begin
  if ColCount > 0 then
  begin
    // Ermitteln, wie breit die Tabelle wrde, wenn alle Spalten genauso
    // breit wren, da alles gerade lesbar ist
    NextSize := 0;
    for Col := 0 to ColCount - 1 do
    begin
      Cols[Col].AutoWidth(Canvas);
      NextSize := NextSize + Cols[Col].Width;
    end;


    Stauchen := NextSize > Breite;
    Stauchbreite := Breite;
    if Stauchen then
    begin // Wenn die Spalten mehr Platz bruchten als verfgbar ist,
          // dann drfen die Spalten, die immer voll Sichtbar sein
          // mssen nicht gestaucht.
      NextSize := 0;
      for Col := 0 to ColCount - 1 do
      begin
        if Cols[Col].BreiteAbschneidbar then
        begin
          NextSize := NextSize + Cols[Col].Width;
        end
        else
        begin
          Stauchbreite := Stauchbreite - Cols[Col].Width;
        end;
      end;
    end;

    // Die Spaltenbreiten so vergrern/verkleiner, da jeder Spalte
    // entsprechend seines breitesten Textes gleich
    // viel Platz zur Verfgung steht wie den anderen Spalten
    // Davon ausgenommen sind Spalten, die in voller Breite sichtbar sein mssen
    if NextSize > 0 then
    begin
      for Col := 0 to ColCount - 1 do
      begin
        if not Stauchen or Cols[Col].BreiteAbschneidbar then
          Cols[Col].Width := trunc(Stauchbreite / NextSize * Cols[Col].Width);
      end;
    end;
  end;
end;

procedure TRakCustomTableData.AutoCellsValueHeihgt(Canvas: TCanvas);
begin
end;

procedure TRakCustomTableData.AutoRowHeights(Canvas: TCanvas; Hoehe: integer);
var
  Row, Y: integer;
begin
  if RowCount > 0 then
  begin
    Y := Hoehe div RowCount;
    if Y < MinRowHeight then
      Y := MinRowHeight
    else if Y > MaxRowHeight then
      Y := MaxRowHeight;
    for Row := 0 to RowCount - 1 do
      Rows[Row].Height := Y;
  end;
end;

procedure TRakCustomTableData.AutoSize(Canvas: TCanvas; Hoehe, Breite: integer);
begin
  // Die Hhe der Zeilen festlegen
  AutoRowHeights(Canvas, Hoehe);

  // Die Hhe der Tabellenwerte so gro machen, da sie
  // von der Hhe her genug Platz haben
  AutoCellsValueHeihgt(Canvas);

  // Die Spaltenbreiten so vergrern/verkleiner, da jeder Spalte
  // entsprechend seines breitesten Wertes gleich
  // viel Platz zur Verfgung steht wie den anderen Spalten
  AutoColWidths(Canvas, Breite);
end;

function TRakCustomTableData.GetCellsObject(Col, Row: integer): TRakTableCell;
begin
  Result := FRows[Row][Col];
end;

procedure TRakCustomTableData.SetCellsObject(Col, Row: integer;
  Value: TRakTableCell);
begin
  FRows[Row][Col] := Value;
end;

procedure TRakCustomTableData.SetAutoRowHeight(Value: TAutoRowHeight);
begin
  FAutoRowHeight := Value;
end;

procedure TRakCustomTableData.SetAutoColumnWidth(Value: TAutoColumnWidth);
begin
  FAutoColumnWidth := Value;
end;

procedure TRakCustomTableData.SetAutoValueHeight(Value: TAutoValueHeight);
var
  Col, Row: integer;
begin
  FAutoValueHeight := Value;
  for Row := 0 to RowCount - 1 do
  begin
    for Col := 0 to ColCount - 1 do
    begin
      CellsObject[Col, Row].AutoValueHeight := AutoValueHeight;
    end;
  end;
end;

procedure TRakCustomTableData.SetAutoValueWidth(Value: TAutoValueWidth);
var
  Col, Row: integer;
begin
  FAutoValueWidth := Value;
  for Row := 0 to RowCount - 1 do
  begin
    for Col := 0 to ColCount - 1 do
    begin
      CellsObject[Col, Row].AutoValueWidth := AutoValueWidth;
    end;
  end;
end;

procedure TRakCustomTableData.SetAutoUmbruch(Value: TAutoUmbruch);
begin
  FAutoUmbruch := Value;
end;

function  TRakCustomTableData.GetColAusrichtung(Id: integer): THorzAusricht;
begin
  Result := FCols[Id].Ausrichtung;
end;

procedure TRakCustomTableData.SetColAusrichtung(Id: integer;
  Value: THorzAusricht);
begin
  FCols[Id].Ausrichtung := Value;
end;

function  TRakCustomTableData.GetColCount: integer;
begin
  Result := FCols.Count;
end;

procedure TRakCustomTableData.SetColCount(Value: integer);
var
  i,t: integer;
begin
  for i := 0 to FRows.Count - 1 do
  begin
    if Value < FRows[i].Count then
    begin
      for t := FRows[i].Count -1 downto Value do
        FRows[i][t].Free;
    end
    else
    begin
      for t := 1 to Value - FRows[i].Count do
        FRows[i].AddNil;
    end;
  end;

  if Value < FCols.Count then
  begin
    for t := FCols.Count -1 downto Value do
      FCols[t].Free;
  end
  else
  begin
    for t := 1 to Value - FCols.Count do
      FCols.Add;
  end;
end;

function  TRakCustomTableData.GetColWidths(Id: integer): integer;
begin
  Result := FCols[Id].Width;
end;

procedure TRakCustomTableData.SetColWidths(Id: integer; Value: integer);
begin
  FCols[Id].Width := Value;
end;

function  TRakCustomTableData.GetRowCount: integer;
begin
  Result := FRows.Count;
end;

procedure TRakCustomTableData.SetRowCount(Value: integer);
var
  i: integer;
begin
  if Value < FRows.Count then
  begin
    for i := FRows.Count -1 downto Value do
      FRows[i].Free;
  end
  else
  begin
    for i := 1 to Value - FRows.Count do
      FRows.Add;
  end;
end;



constructor TRakTableCellLine.Create(Parent: TRakTableCell);
begin
  inherited Create;

  FPen := TPen.Create;
  FPen.OnChange := PenChange;
  FParentCell := Parent;
  FTyp := ltNone;
end;

destructor TRakTableCellLine.Destroy;
begin
  FPen.Free;

  inherited Destroy;
end;

procedure TRakTableCellLine.Assign(Source: TPersistent);
var
  Src: TRakTableCellLine;
begin
  if Source is TRakTableCellLine then
  begin
    Src := TRakTableCellLine(Source);

    FPen.Assign(Src.FPen);
    FTyp := Src.FTyp;
  end;
end;

procedure TRakTableCellLine.PenChange(Sender: TObject);
begin
  if Assigned(FParentCell) then
    FParentCell.GraphicObjectChanged(Sender);
end;

procedure TRakTableCellLine.SetTyp(Value: TRakTableCellLineTyp);
begin
  if FTyp <> Value then
  begin
    FTyp := Value;
    Changed;
  end;
end;

procedure TRakTableCellLine.SetVisible(Value: boolean);
begin
  if FVisible <> Value then
  begin
    FVisible := Value;
    Changed;
  end;
end;

procedure TRakTableCellLine.Changed;
begin
  if Assigned(FParentCell) then
    FParentCell.LineChanged(self);
end;

procedure TRakTableCellLine.Paint(Canvas: TCanvas; ClientRect: TRect;
  Seite: integer);
begin
  if Typ <> ltNone then
  begin
    with Canvas do
    begin
      Pen.Assign(FPen);
      case Seite of
      0:
        begin
          case Typ of
          ltSingle:
            begin
              MoveTo(ClientRect.Left, ClientRect.Top);
              LineTo(ClientRect.Left, ClientRect.Bottom+1);
            end;
          end;
        end;
      1:
        begin
          case Typ of
          ltSingle:
            begin
              MoveTo(ClientRect.Left, ClientRect.Top);
              LineTo(ClientRect.Right+1, ClientRect.Top);
            end;
          end;
        end;
      2:
        begin
          case Typ of
          ltSingle:
            begin
              MoveTo(ClientRect.Right, ClientRect.Top);
              LineTo(ClientRect.Right, ClientRect.Bottom+1);
            end;
          end;
        end;
      3:
        begin
          case Typ of
          ltSingle:
            begin
              MoveTo(ClientRect.Left, ClientRect.Bottom);
              LineTo(ClientRect.Right+1, ClientRect.Bottom);
            end;
          end;
        end;
      end
    end;
  end;
end;


procedure TRakTableCellLine.LoadAsStream(Reader: TReader);
var
  iWert: integer;
begin
  iWert := Reader.ReadInteger;
  if iWert > 1 then
    raise TRakTableVersionException.Create('Version 1');

  FPen.Width := Reader.ReadInteger;
  FPen.Color := Reader.ReadInteger;
  FPen.Style := TPenStyle(Reader.ReadInteger);
  FPen.Mode := TPenMode(Reader.ReadInteger);

  FTyp := TRakTableCellLineTyp(Reader.ReadInteger);
  FVisible := Reader.ReadBoolean;
end;

procedure TRakTableCellLine.SaveAsStream(Writer: TWriter);
begin
  Writer.WriteInteger(1);

  Writer.WriteInteger(FPen.Width);
  Writer.WriteInteger(FPen.Color);
  Writer.WriteInteger(Ord(FPen.Style));
  Writer.WriteInteger(Ord(FPen.Mode));

  Writer.WriteInteger(Ord(FTyp));
  Writer.WriteBoolean(FVisible);
end;



constructor TRakTableCell.Create;
begin
  inherited Create;

  FBrush := TBrush.Create;
  FBrush.OnChange := GraphicObjectChanged;

  FLineLeft := TRakTableCellLine.Create(self);
  FLineTop := TRakTableCellLine.Create(self);
  FLineRight := TRakTableCellLine.Create(self);
  FLineBottom := TRakTableCellLine.Create(self);

  FHorzAusricht := haLinks;
  FHorzOffset := 0;
  FLastChange := clcElse;
end;

destructor TRakTableCell.Destroy;
begin
  FBrush.Free;

  FLineLeft.Free;
  FLineTop.Free;
  FLineRight.Free;
  FLineBottom.Free;

  inherited Destroy;
end;

procedure TRakTableCell.Assign(Source: TRakCollectionItem);
var
  Src: TRakTableCell;
begin
  inherited Assign(Source);

  if Source is TRakTableCell then
  begin
    Src := TRakTableCell(Source);

    FBrush.Assign(Src.FBrush);
    FLineLeft.Assign(Src.FLineLeft);
    FLineTop.Assign(Src.FLineTop);
    FLineRight.Assign(Src.FLineRight);
    FLineBottom.Assign(Src.FLineBottom);
  end;
end;

function TRakTableCell.GetValueHeight(Canvas: TCanvas): integer;
begin
  Result := 0;
end;

function TRakTableCell.AutoFormat(Canvas: TCanvas; MaxHeightFor_avhWidth: integer): boolean;
begin
  Result := false;
end;


procedure TRakTableCell.GraphicObjectChanged(Sender: TObject);
begin
  if Sender is TBrush then
    FLastChange := clcBrush
  else if Sender is TPen then
    FLastChange := clcPen
  else if Sender is TFont then
    FLastChange := clcFont
  else
    FLastChange := clcElse;

  Changed(self);
end;

procedure TRakTableCell.LineChanged(Line: TRakTableCellLine);
begin
   FLastChange := clcLine;
  Changed(self);
end;

procedure TRakTableCell.SetHorzAusricht(Value: THorzAusricht);
begin
  if FHorzAusricht <> Value then
  begin
    if FHorzAusricht = haKomma then
      raise Exception.Create('haKomma ist bei einer Zelle kein gltiger Wer fr HorzAusricht');
    FHorzAusricht := Value;

    FLastChange := clcAusrichtung;
    Changed(self);
  end;
end;

procedure TRakTableCell.SetHorzOffset(Value: integer);
begin
  if FHorzOffset <> Value then
  begin
    FHorzOffset := Value;

    if HorzAusricht in [haLinksOffset, haRechtsOffset] then
    begin
      FLastChange := clcOffset;
      Changed(self);
    end;
  end;
end;

function TRakTableCell.GetParentCol: TRakTableDataCollumn;
begin
  Result := ParentTable.Cols[Index];
end;

function TRakTableCell.GetParentRow: TRakTableDataRow;
begin
  Result := TRakTableDataRow(FCollection);
end;

function TRakTableCell.GetParentTable: TRakCustomTableData;
begin
  Result := ParentRow.FTable;
end;

function TRakTableCell.GetValueNachKommaWidth(Canvas: TCanvas): integer;
begin
  Result := 0;
end;

function TRakTableCell.GetValueWidth(Canvas: TCanvas): integer;
begin
  Result := 0;
end;

function TRakTableCell.GetValueVorKommaWidth(Canvas: TCanvas): integer;
begin
  Result := 0;
end;

procedure TRakTableCell.PaintBackground(Canvas: TCanvas; ClientRect: TRect;
  Drehung: integer);
var
  Points: array[0..3] of TPoint;
  X, Y: integer;
begin
  with Canvas do
  begin
    Brush.Assign(FBrush);
    Pen.Style := psClear;
    Pen.Mode := pmCopy;

    Points[0] := Point(ClientRect.Left, ClientRect.Top);

    X := ClientRect.Right;
    Y := ClientRect.Top;
{$IFDEF USE_WYSIWYGFONT}
    Drehe(X, Y, ClientRect.Left, ClientRect.Top, Drehung);
{$ENDIF}
    Points[1] := Point(X, Y);

    X := ClientRect.Right;
    Y := ClientRect.Bottom;
{$IFDEF USE_WYSIWYGFONT}
    Drehe(X, Y, ClientRect.Left, ClientRect.Top, Drehung);
{$ENDIF}
    Points[2] := Point(X, Y);

    X := ClientRect.Left;
    Y := ClientRect.Bottom;
{$IFDEF USE_WYSIWYGFONT}
    Drehe(X, Y, ClientRect.Left, ClientRect.Top, Drehung);
{$ENDIF}
    Points[3] := Point(X, Y);

    Polygon(Points);
{    Rectangle(ClientRect.Left, ClientRect.Top, ClientRect.Right+1,
      ClientRect.Bottom+1);}
  end;
end;

procedure TRakTableCell.LoadAsStream(Reader: TReader);
var
  iWert: integer;
begin
  iWert := Reader.ReadInteger;
  if iWert > 1 then
    raise TRakTableVersionException.Create('Version 1');

  MaxHeight := Reader.ReadInteger;
{$IFDEF USE_WYSIWYGFONT}
  MaxWidth := Reader.ReadInteger;
{$ENDIF}

  FHorzAusricht := THorzAusricht(Reader.ReadInteger);
  FHorzOffset := Reader.ReadInteger;
  FLastChange := TCellLastChange(Reader.ReadInteger);

  FLineLeft.LoadAsStream(Reader);
  FLineTop.LoadAsStream(Reader);
  FLineRight.LoadAsStream(Reader);
  FLineBottom.LoadAsStream(Reader);
end;

procedure TRakTableCell.SaveAsStream(Writer: TWriter);
begin
  Writer.WriteInteger(1);

  Writer.WriteInteger(MaxHeight);
{$IFDEF USE_WYSIWYGFONT}
  Writer.WriteInteger(MaxWidth);
{$ENDIF}

  Writer.WriteInteger(Ord(FHorzAusricht));
  Writer.WriteInteger(FHorzOffset);
  Writer.WriteInteger(Ord(FLastChange));

  FLineLeft.SaveAsStream(Writer);
  FLineTop.SaveAsStream(Writer);
  FLineRight.SaveAsStream(Writer);
  FLineBottom.SaveAsStream(Writer);
end;

procedure TRakTableCell.Paint(Canvas: TCanvas; ClientRect: TRect;
  Drehung: integer);
begin
  FLineLeft.Paint(Canvas, ClientRect, 0);
  FLineTop.Paint(Canvas, ClientRect, 1);
  FLineRight.Paint(Canvas, ClientRect, 2);
  FLineBottom.Paint(Canvas, ClientRect, 3);
end;



constructor TRakTableDataCollumn.Create;
begin
  inherited Create;

  FAusrichtung := haLinks;
  FBreiteAbschneidbar := true;
  FTable := nil;
  FVisible := true;
  FWidth := 10;
  FLine := ltNone;
end;

procedure TRakTableDataCollumn.Assign(Source: TRakCollectionItem);
var
  Src: TRakTableDataCollumn;
begin
  inherited Assign(Source);

  if Source is TRakTableDataCollumn then
  begin
    Src := TRakTableDataCollumn(Source);

    Ausrichtung := Src.FAusrichtung;
    Width := Src.FWidth;
    Line := Src.FLine;
  end;
end;

procedure TRakTableDataCollumn.AutoWidth(Canvas: TCanvas);
var
  X, DX, DX2: integer;
  Row: integer;
begin
  if Assigned(FTable) then
  begin // Wenn die Spalte zu einer Tabelle gehrt
    // Fr jede Zelle die Breite ermitteln, welche bentigt wird, wenn
    // die Zelle die komplette Breite des Inhalts darstellen mchte
    // Die grte Breite wird sich gemerkt und die Spalte dann
    // entsprechend breit gemacht

    // DX enthlt die bisher grte Breite der Zellen
    // Wenn 'Ausrichtung = haKomma' dann enthlt DX nur die Breite
    // des Teils vor dem Komma
    DX := 0;
    // Wenn 'Ausrichtung = haKomma' dann enthlt DX2 die Breite
    // des Teils nach dem Komma und des Kommas
    // Wenn 'Ausrichtung <> haKomma', dann wird DX2 nicht verndert
    DX2 := 0;

    for Row := 0 to FTable.RowCount - 1 do
    begin // Alle Zellen durchgehen
      if Ausrichtung = haKomma then
      begin // Wenn der Inhalt der Zellen nach einem bestimmten
            // Punkt ausgerichtet werden mu
            // (z.B.: Die Zellen enthalten Kommazahlen und die Zahlen
            // sollen alle so Ausgerichtet sein, da die Kommas
            // genau bereinander stehen, egal wieviele Nach- oder
            // Vorkommastellen die einzelnen Zellen haben
        X := Cells[Row].GetValueVorKommaWidth(Canvas);
        if X > DX then
          DX := X;
        X := Cells[Row].GetValueNachKommaWidth(Canvas);
        if X > DX2 then
          DX2 := X;
        // Hier wird sich die Breite der 'Nachkommagre' gemerkt,
        // damit sie weiter unten, wo sie bentigt wird, nicht nocheinmal
        // ausgerechenet werden mu
        Cells[Row].HorzOffset := X;
      end
      else
      begin
        X := Cells[Row].GetValueWidth(Canvas);
        if X > DX then
          DX := X;
      end;
    end;

    // Die Breite der Spalte so gro machen, damit sich alle Zellen komplett
    // anzeigen knnen
    Width := DX + DX2;

    if Ausrichtung = haKomma then
    begin // Wenn der Inhalt der Zellen nach einem bestimmten
          // Punkt ausgerichtet werden mu
      // Dann wird hier dafr gesorgt, da die Zellen
      // ihren Inhalt auf die entsprechende Position verschieben
      for Row := 0 to FTable.RowCount - 1 do
      begin
        Cells[Row].HorzAusricht := haRechtsOffset;
        Cells[Row].HorzOffset := DX2 - Cells[Row].HorzOffset;
      end;
    end;
  end;
end;

procedure TRakTableDataCollumn.LoadAsStream(Reader: TReader);
var
  iWert: integer;
  i: integer;
begin
  iWert := Reader.ReadInteger;
  if iWert > 1 then
    raise TRakTableVersionException.Create('Version 1');

  Ausrichtung := THorzAusricht(Reader.ReadInteger);
  BreiteAbschneidbar := Reader.ReadBoolean;
  ColName := Reader.ReadString;
  Width := Reader.ReadInteger;
  Line := TRakTableCellLineTyp(Reader.ReadInteger);

  for i := 0 to FTable.RowCount - 1 do
    Cells[i].LoadAsStream(Reader);
end;

procedure TRakTableDataCollumn.SaveAsStream(Writer: TWriter);
var
  i: integer;
begin
  Writer.WriteInteger(1);

  Writer.WriteInteger(Ord(Ausrichtung));
  Writer.WriteBoolean(BreiteAbschneidbar);
  Writer.WriteString(ColName);
  Writer.WriteInteger(Width);
  Writer.WriteInteger(Ord(Line));

  for i := 0 to FTable.RowCount - 1 do
    Cells[i].SaveAsStream(Writer);
end;



procedure TRakTableDataCollumn.SetAusrichtung(Value: THorzAusricht);
var
  i: integer;
begin
  FAusrichtung := Value;

  if Assigned(FTable) then
  begin
    if Value = haKomma then
      Value := haRechts;

    for i := 0 to FTable.RowCount - 1 do
    begin
      FTable.CellsObject[Index, i].HorzAusricht := Value;
    end;
  end;

  Changed(self);
end;

function TRakTableDataCollumn.GetCells(Row: integer): TRakTableCell;
begin
  if Assigned(FTable) then
  begin
    Result := FTable.GetCellsObject(Index, Row);
  end
  else
  begin
    Result := nil;
  end;
end;

procedure TRakTableDataCollumn.SetCollection(Value: TRakCollection);
begin
  if (Collection <> Value) then
  begin
    IgnoreChange := true;
    try
      inherited SetCollection(Value);

      if (Collection <> nil) and (Collection is TRakTableDataCollumns) then
      begin
        FTable := (Collection as TRakTableDataCollumns).FTable;
      end;
    finally
      IgnoreChange := false;
    end;
    Changed(self);
  end;
end;

procedure TRakTableDataCollumn.SetColName(Value: string);
begin
  FColName := Value;
end;

procedure TRakTableDataCollumn.SetLine(Value: TRakTableCellLineTyp);
var
  i: integer;
begin
  FLine := Value;

  if Assigned(FTable) then
  begin
    for i := 0 to FTable.RowCount - 1 do
    begin
      FTable.CellsObject[Index, i].LineRight.Typ := Value;
    end;
  end;

  Changed(self);
end;

function TRakTableDataCollumn.GetVisible: boolean;
begin
  Result := FVisible;
end;

procedure TRakTableDataCollumn.SetVisible(Value: boolean);
begin
  if Value <> Visible then
  begin
    FVisible := Value;
    Changed(self);
  end;
end;

function TRakTableDataCollumn.GetVisibleIndex: integer; 
begin
  Result := TRakTableDataCollumns(Collection).GetVisibleIndex(self);
end;

procedure TRakTableDataCollumn.SetWidth(Value: integer);
begin
  FWidth := Value;
  Changed(self);
end;

function TRakTableDataCollumns.GetItems(Index: integer): TRakTableDataCollumn;
begin
  Result := TRakTableDataCollumn(inherited GetItems(Index));
end;

procedure TRakTableDataCollumns.SetItems(Index: Integer;
  Value: TRakTableDataCollumn);
begin
  inherited SetItems(Index, Value);
end;

function TRakTableDataCollumns.GetVisibleItems(Idx: Integer): TRakTableDataCollumn;
var
  i, t: integer;
begin
  Result := nil;
  
  t := 0;
  for i := 0 to Count - 1 do
  begin
    if Items[i].Visible then
    begin
      if t = Idx then
      begin
        Result := Items[i];
        break;
      end;
      inc(t);
    end;
  end;
end;

function TRakTableDataCollumns.Add: TRakTableDataCollumn;
begin
  Result := AddTyped(TRakTableDataCollumn);
end;

function TRakTableDataCollumns.AddTyped(Typ: TRakTableDataCollumnClass):
  TRakTableDataCollumn;
begin
  Result := TRakTableDataCollumn(inherited Add(Typ));
  Result.FTable := FTable;
end;

function TRakTableDataCollumns.CountVisible: integer;
var
  i: integer;
begin
  Result := 0;
  for i := 0 to Count - 1 do
  begin
    if Items[i].Visible then
      inc(Result);
  end;
end;

function TRakTableDataCollumns.GetVisibleIndex(Item: TRakTableDataCollumn): integer;
var
  i, t: integer;
begin
  Result := -1;
  t := 0;
  for i := 0 to Count -1 do
  begin
    if Items[i].Visible then
    begin
      if Items[i] = Item then
      begin
        Result := t;
        break;
      end;
      inc(t);
    end;
  end;
end;

function TRakTableDataCollumns.ItemByName(ColName: string): TRakTableDataCollumn;
var
  i: integer;
begin
  Result := nil;
  i := 0;
  while (Result = nil) and (i < Count) do
  begin
    if CompareText(Items[i].ColName, ColName) = 0 then
      Result := Items[i];
    inc(i);
  end;
end;

procedure TRakTableDataCollumns.LoadAsStream(Reader: TReader);
var
  iWert: integer;
  i: integer;
begin
  iWert := Reader.ReadInteger;
  if iWert > 1 then
    raise TRakTableVersionException.Create('Version 1');

  Count := Reader.ReadInteger;
  for i := 0 to Count - 1 do
    Items[i].LoadAsStream(Reader);
end;

procedure TRakTableDataCollumns.SaveAsStream(Writer: TWriter);
var
  i: integer;
begin
  Writer.WriteInteger(1);

  Writer.WriteInteger(Count);
  for i := 0 to Count - 1 do
    Items[i].SaveAsStream(Writer);
end;


constructor TRakTableDataRow.Create;
begin
  inherited Create;

  FHeight :=10;
  FLine := ltNone;
  FTable := nil;
end;

function TRakTableDataRow.Add(
  TheClass: TRakCollectionItemClass): TRakCollectionItem;
begin
  Result := inherited Add(TheClass);
  if Result is TRakTableCell then
  begin
    with Result as TRakTableCell do
    begin
//      FTable := self.FTable;
      if FTable.ColCount > Index then
      begin
        if ParentCol.Ausrichtung = haKomma then
          HorzAusricht := haRechts
        else
          HorzAusricht := ParentCol.Ausrichtung;
        LineBottom.Typ := Line;
        LineRight.Typ := ParentCol.Line;
      end;
      AutoValueHeight := FTable.AutoValueHeight;
      AutoValueWidth := FTable.AutoValueWidth;
    end;
  end;
end;

procedure TRakTableDataRow.Assign(Source: TRakCollectionItem);
begin
  inherited Assign(Source);

  if Source is TRakTableDataRow then
  begin
    StateAdd(csAssign);
    try
      Line := (Source as TRakTableDataRow).Line;
      Height := (Source as TRakTableDataRow).Height;
    finally
      StateDelete(csAssign);
    end;
    Changed(self);
  end;
end;

procedure TRakTableDataRow.LoadAsStream(Reader: TReader);
var
  iWert: integer;
  i: integer;
begin
  iWert := Reader.ReadInteger;
  if iWert > 1 then
    raise TRakTableVersionException.Create('Version 1');

  Height := Reader.ReadInteger;
  Line := TRakTableCellLineTyp(Reader.ReadInteger);

  for i := 0 to FTable.ColCount - 1 do
    Cells[i].LoadAsStream(Reader);
end;

procedure TRakTableDataRow.SaveAsStream(Writer: TWriter);
var
  i: integer;
begin
  Writer.WriteInteger(1);

  Writer.WriteInteger(Height);
  Writer.WriteInteger(Ord(Line));

  for i := 0 to FTable.ColCount - 1 do
    Cells[i].SaveAsStream(Writer);
end;


procedure TRakTableDataRow.SetCollection(Value: TRakCollection);
begin
  if (Collection <> Value) then
  begin
    IgnoreChange := true;
    try
      inherited SetCollection(Value);

      if (Collection <> nil) and (Collection is TRakTableRows) then
      begin
        FTable := (Collection as TRakTableRows).FTable;
      end;
    finally
      IgnoreChange := false;
    end;
    Changed(self);
  end;
end;

procedure TRakTableDataRow.SetHeight(Value: integer);
begin
  if Value <> Height then
  begin
    FHeight := Value;
    Changed(self);
  end;
end;

function TRakTableDataRow.GetCells(Index: Integer): TRakTableCell;
begin
  Result := TRakTableCell(inherited GetItems(Index));
end;

procedure TRakTableDataRow.SetCells(Index: Integer; Value: TRakTableCell);
begin
  inherited SetItems(Index, Value);
end;
{
function TRakTableDataRow.Add: TRakTableCell;
begin
  Result := TRakTableCell(inherited Add(TRakTableCell));
end;
}
procedure TRakTableDataRow.SetLine(Value: TRakTableCellLineTyp);
var
  i: integer;
begin
  FLine := Value;

  if Assigned(FTable) then
  begin
    for i := 0 to FTable.ColCount - 1 do
    begin
      FTable.CellsObject[i, Index].LineBottom.Typ := Value;
    end;
  end;

  Changed(self);
end;



function TRakTableRows.GetItems(Index: Integer): TRakTableDataRow;
begin
  Result := TRakTableDataRow(inherited GetItems(Index));
end;

procedure TRakTableRows.SetItems(Index: Integer; Value: TRakTableDataRow);
begin
  inherited SetItems(Index, Value);
end;

function TRakTableRows.Add: TRakTableDataRow;
begin
  Result := TRakTableDataRow(inherited Add(TRakTableDataRow));
  Result.FTable := FTable;
end;

procedure TRakTableRows.LoadAsStream(Reader: TReader);
var
  iWert: integer;
  i: integer;
begin
  iWert := Reader.ReadInteger;
  if iWert > 1 then
    raise TRakTableVersionException.Create('Version 1');

  Count := Reader.ReadInteger;
  for i := 0 to Count - 1 do
    Items[i].LoadAsStream(Reader);
end;

procedure TRakTableRows.SaveAsStream(Writer: TWriter);
var
  i: integer;
begin
  Writer.WriteInteger(1);

  Writer.WriteInteger(Count);
  for i := 0 to Count - 1 do
    Items[i].SaveAsStream(Writer);
end;


constructor TRakStringCell.Create;
begin
  inherited Create;

  FFont := TFont.Create;
  FFont.OnChange := GraphicObjectChanged;

  FFuncRec := TFuncRec.Create;
  FFuncRec.GetTextWidth := TextWidth;
  FFuncRec.GetTextHeight := TextHeight;
  FFuncRec.GetFontSize := GetFontSize;
  FFuncRec.SetFontSize := SetFontSize;

{$IFDEF USE_WYSIWYGFONT}
  FFuncRec.GetFontWidth := GetFontWidth;
  FFuncRec.SetFontWidth := SetFontWidth;

  FUseWYSIWYGFont := true;
  FWYSIWYGFont := TRakWYSIWYGFont.Create;
{$ENDIF}
end;

destructor TRakStringCell.Destroy;
begin
  FFont.Free;
  FFuncRec.Free;

{$IFDEF USE_WYSIWYGFONT}
  FWYSIWYGFont.Free;
{$ENDIF}

  inherited Destroy;
end;

function TRakStringCell.AutoFormat(Canvas: TCanvas; MaxHeightFor_avhWidth: integer): boolean;
  function CountUmbrueche: integer;
    procedure ScanWort(Zeile: string; var Rest, Scanned: string);
    begin
      Rest := Zeile;
      Scanned := '';
      // Zuerst die fhrenden Leerzeichen scannen
      while Pos(' ', Rest) > 0 do
      begin
        Rest := Copy(Rest, 2, length(Rest) - 1);
        Scanned := Scanned + ' ';
      end;
      // Jetzt wird das eigentliche Wort gescannt
      while (length(Rest) > 0) and (Rest[1] <> ' ') do
      begin
        Scanned := Scanned + Rest[1];
        Rest := Copy(Rest, 2, length(Rest) - 1);
      end;
    end;
  var
    Zeile, Rest, Anfang, Scanned: string;
  begin // CountUmbrueche
    // Feststellen, wie oft umgebrochen werden mte bei der aktuellen Schriftgre
    Zeile := Value;
    Result := 0;
    repeat
      // In Anfang wird der Teil der Zeile gespeichert, der in die Spalte pat
      // Anfang enthlt dann nur ganze Wrter
      Anfang := '';
      ScanWort(Zeile, Rest, Scanned);
      while (Scanned <> '')
        and (TextWidth(Anfang + Scanned) <= ParentCol.Width) do
      begin
        Anfang := Anfang + Scanned;
        Zeile := Rest;
        ScanWort(Zeile, Rest, Scanned);
      end;

      if Anfang = '' then
      begin // Wenn selbst das erste Wort zu breit fr die Spalte ist
        if Scanned <> '' then
        begin // Wenn es berhaupt ein erstes Wort gibt
          // Das erste Wort Buchstabe fr Buchstabe krzen, bis
          // es in die Spalte pat
          while (Scanned <> '')
            and (TextWidth(Scanned) > ParentCol.Width) do
          begin
            SetLength(Scanned, length(Scanned) - 1);
          end;
          if Scanned <> '' then
          begin // Wenn mindestens fr den ersten Buchstaben in der Spalte
                // Platz genug ist
            // Merken, welcher Teil des Wortes in die Spalte pat
            Anfang := Scanned;
            // und die Zeile um den entsprechenden Teil krzen
            Zeile := Copy(Zeile, length(Scanned) + 1, length(Zeile));
          end;
        end;
      end;
      // Fhrende Leerzeichen werden nicht angezeigt
      while Pos(' ', Zeile) = 1 do
        Zeile := Copy(Zeile, 2, length(Zeile) - 1);
      // Merken, da eine weitere Zeile bentigt wird
      Result := Result + 1;
    // Nur wenn etwas in die Spalte gepat hat und es noch etwas fr eine
    // nchste Zeile gbe, wird versucht festzustellen, was in die nchste
    // Zeile pat
    until (Anfang = '') or (Zeile = '');
  end; // CountUmbrueche

var
  OrgFontSize: integer;
  ExpectedCount: integer;
begin // AutoFormat
  Result := false;
  FCanvas := Canvas;

  if ParentTable <> nil then
  begin // Wenn die Zelle zu einer Tabelle gehrt
    if ParentRow <> nil then
    begin // Wenn die Zelle zu einer Reihe gehrt
      // Dann mu der Wert von 'AutoValueHeight' bercksichtigt werden
      case AutoValueHeight of
      avhFix:; // Die Hhe der Schrift soll nicht verndert werden
      avhAuto: // Die Hhe der Schrift soll so hoch sein, da sie genau reinpat
        FitFontSize(FFuncRec, ParentRow.Height);
      avhMax: // Die Hhe der Schrift soll so hoch sein, da sie genau reinpat, wobei sie eine maximale Hhe nicht berschreiten darf
        if ParentRow.Height > MaxHeight then
        begin // Wenn die Reihe hher ist, als die Schrift maximal sein darf
          FitFontSize(FFuncRec, MaxHeight);
        end
        else
        begin
          FitFontSize(FFuncRec, ParentRow.Height);
        end;
      end;
    end;

    // Merken, wieviele Pixel die Schrift in der Hhe bentigt
    NeededHeight := TextHeight(Value);

    // Merken, wieviele Pixel der Text in der Breite bentigt
    NeededWidth := TextWidth(Value);
    // Merken, da kein Umbruch erfolgen soll
    ZeilenCount := 1;

    if ParentCol <> nil then
    begin // Wenn die Zelle zu einer Spalte gehrt
      // Dann mu auch die Textbreite bercksichtigt werden
      if AutoValueHeight = avhWidth then
      begin // Wenn sich die Hhe der Schrift nach der Breite richten soll
        // Dann wird die Schriftgre so eingestellt, da der Text die Breite der Spalte voll ausnutzt
        FitFontSizeByWidth(FFuncRec, ParentCol.Width, Value);

        // Merken, wieviele Pixel die Schrift in der Hhe bentigt
        NeededHeight := TextHeight(Value);

        if NeededHeight > MaxHeightFor_avhWidth then
        begin
          FitFontSize(FFuncRec, MaxHeightFor_avhWidth);
          // Merken, wieviele Pixel die Schrift in der Hhe bentigt
          NeededHeight := TextHeight(Value);
        end;

        // Merken, wieviele Pixel der Text in der Breite bentigt
        NeededWidth := TextWidth(Value);
      end
      else
      begin // Wenn sich die Hhe der Schrift nicht nach der Breite richten soll
        // Dann mu die Breite der Schrift ermittelt werden
        case AutoValueWidth of
        avwFix:; // Wenn die Breite nicht verndert werden soll
{$IFDEF USE_WYSIWYGFONT}
        avwAuto: // Wenn die Schrift so breit wie mglich gemacht werden soll
          // Dann wird die Schriftbreite an die Breite der Spalte angepat
          FitFontWidth(FFuncRec, ParentCol.Width, Value);
        avwMax: // Wenn die Schrift so breit wie mglich gemacht werden soll, wobei sie eine maximale Breite nicht berschreiten darf
          if ParentCol.Width > MaxWidth then
          begin // Wenn die Spalte breiter ist, als die Schrift maximal sein darf
            FitFontWidth(FFuncRec, MaxWidth, Value);
          end
          else
          begin
            FitFontWidth(FFuncRec, ParentCol.Width, Value);
          end;
        avwStauchen: // Wenn die Schrift gestaucht werden soll, wenn die Spalte nicht breit genug ist fr den Text
          begin
            FontWidth := 0;
            if ParentCol.Width < TextWidth(Value) then
              FitFontWidth(FFuncRec, ParentCol.Width, Value);
          end;
{$ENDIF}
        avwUmbruch: // Wenn der Text umgebrochen werden soll, wenn die Spalte nicht breit genug ist fr den Text
          begin
            if AutoUmbruch = auFix then
            begin // Wenn bei einem ntigen Umbruch die Hhe der Reihe nicht verndert werden darf
              // Die Ausgangsgre der Schrift merken
              OrgFontSize := FontSize;

              // Davon ausgehen, da man nur eine Zeile braucht
              ExpectedCount := 1;
              // Feststellen, wieviele Zeilen man bei der aktuellen Schriftgre bruchte
              ZeilenCount := CountUmbrueche;
              while ExpectedCount < ZeilenCount do
              begin // Solange man von weniger Zeilen ausgegangen ist, als bei der aktuellen Schriftgre bentigt wrden
                // Davon ausgehen, da eine weitere Zeile bentigt wird
                inc(ExpectedCount);
                // Schriftgre auf den Wert setzen, mit dem alle erwarteten Zeilen in die Zelle passen wrden
                FontSize := OrgFontSize div ExpectedCount;
                // Feststellen, wieviele Zeilen man bei der aktuellen Schriftgre bruchte
                ZeilenCount := CountUmbrueche;
              end;

              if ZeilenCount < ExpectedCount then
              begin // Wenn man bei der aktuellen Schriftgre weniger Zeilen bruchte, als erwartet
                // Dann wird die Schrift solange vergrert bis sich die Anzahl der bentigten Zeilen erhht
                repeat
                  FontSize := FontSize + 1;
                until CountUmbrueche > ZeilenCount;
                FontSize := FontSize - 1;
              end;
            end
            else
            begin // Wenn bei einem ntigen Umbruch die Hhe der Reihe sich an die Hhe des Textes anpat
              ZeilenCount := CountUmbrueche;
            end;
            // Merken, wieviele Pixel die Schrift in der Hhe bentigt
            NeededHeight := TextHeight(Value) * ZeilenCount;
            NeededWidth := ParentCol.Width;
          end;
{$IFDEF USE_WYSIWYGFONT}
        avwStauchenUmbruch:;
{$ENDIF}
        end;
      end;
    end;
  end;
end;

function TRakStringCell.GetValueNachKommaWidth(Canvas: TCanvas): integer;
var
  Wo: integer;
begin
  FCanvas := Canvas;

  Wo := Pos(',', Value);
  if Wo > 0 then
    Result := TextWidth(Copy(Value, Wo, length(Value)))
  else
    Result := 0;
end;

function TRakStringCell.GetValueWidth(Canvas: TCanvas): integer;
begin
  FCanvas := Canvas;
  Result := TextWidth(Value);
end;

function TRakStringCell.GetValueHeight(Canvas: TCanvas): integer;
begin
  FCanvas := Canvas;
  Result := TextHeight(Value);
end;

function TRakStringCell.GetValueVorKommaWidth(Canvas: TCanvas): integer;
var
  Wo: integer;
begin
  FCanvas := Canvas;
  Wo := Pos(',', Value);
  if Wo > 0 then
    Result := TextWidth(Copy(Value, 1, Wo))
  else
    Result := TextWidth(Value);
end;

procedure TRakStringCell.Paint(Canvas: TCanvas; ClientRect: TRect;
  Drehung: integer);
var
{$IFDEF USE_WYSIWYGFONT}
  DrawData: TDrawData;
  Y: integer;
{$ENDIF}
  ARect: TRect;
  X: integer;
  Zeile: string;
  lBreite: integer;
begin
  FCanvas := Canvas;

  inherited Paint(Canvas, ClientRect, Drehung);

  with Canvas, ClientRect do
  begin
    Brush.Style := bsClear;
    Font.Assign(FFont);

    Zeile := Value;

    if OnlyWholeChars then
    begin
      lBreite := ClientRect.Right - ClientRect.Left + 1;
      while (Zeile <> '')
            and
            (  (  not (HorzAusricht in [haLinksOffset, haRechtsOffset])
                and
                (self.TextWidth(Zeile) > lBreite)
              )
              or
              (  (HorzAusricht in [haLinksOffset, haRechtsOffset])
                and
                (self.TextWidth(Zeile) + HorzOffset > lBreite)
              )
            ) do
      begin
        if HorzAusricht in [haRechts, haRechtsOffset] then
          Zeile := Copy(Zeile, 2, length(Zeile) - 1)
        else
          SetLength(Zeile, length(Zeile) - 1);
      end;
    end;

    ARect := ClientRect;

{$IFDEF USE_WYSIWYGFONT}
    if UseWYSIWYGFont then
    begin
      inc(ARect.Bottom);
      inc(ARect.Right);
      WYSIWYGFont.Font.Angle := Drehung;

      case HorzAusricht of
        haMittig:        X := (Right + Left - self.TextWidth(Zeile)) div 2;
        haRechts:       X := Right - self.TextWidth(Zeile);
        haLinksOffset:  X := Left + HorzOffset;
        haRechtsOffset: X := Right - self.TextWidth(Zeile) - HorzOffset;
        else             X := Left;
      end;
      Y := Top;
      Drehe(X, Y, Left, Top, Drehung);

      DrawData := TDrawData.Create;
      try
        DrawData.DoClip := DoClip;
        DrawData.AClipRect := ARect;
        DrawData.ClipWinkel := Drehung;
        DrawData.Left := X;
        DrawData.Top := Y;
        DrawData.Winkel := Drehung;
        DrawData.Text := Zeile;

        WYSIWYGFont.DrawText(Canvas, DrawData);
      finally
        DrawData.Free;
      end;
    end
    else
{$ENDIF}
    begin
      case HorzAusricht of
        haMittig:        X := (Right + Left - self.TextWidth(Zeile)) div 2;
        haRechts:       X := Right - 2 - self.TextWidth(Zeile);
        haLinksOffset:  X := Left + 2 + HorzOffset;
        haRechtsOffset: X := Right - 2 - self.TextWidth(Zeile) - HorzOffset;
        else             X := Left + 2;
      end;

      if GetDevicecaps(Handle, PLANES) = 1 then
        Font.Color := clBlack;

      ExtTextOut(Canvas.Handle, X, Top, ETO_CLIPPED,
        @ARect, PChar(Zeile), Length(Zeile), nil);
    end;
  end;
end;

procedure TRakStringCell.LoadAsStream(Reader: TReader);
var
  iWert: integer;
begin
  inherited LoadAsStream(Reader);

  iWert := Reader.ReadInteger;
  if iWert > 1 then
    raise TRakTableVersionException.Create('Version 1');

  Font.Name := Reader.ReadString;
  Font.Size := Reader.ReadInteger;
  Font.Color := Reader.ReadInteger;

  FOnlyWholeChars := Reader.ReadBoolean;

{$IFDEF USE_WYSIWYGFONT}
  FUseWYSIWYGFont := Reader.ReadBoolean;
  WYSIWYGFont.Font.Angle := Reader.ReadInteger;
  WYSIWYGFont.Font.Color := Reader.ReadInteger;
  WYSIWYGFont.Font.Name := Reader.ReadString;
  WYSIWYGFont.Font.Pitch := TFontPitch(Reader.ReadInteger);
  WYSIWYGFont.Font.Size := Reader.ReadInteger;
  WYSIWYGFont.Font.Width := Reader.ReadInteger;
{$ENDIF}
end;

procedure TRakStringCell.SaveAsStream(Writer: TWriter);
begin
  inherited SaveAsStream(Writer);

  Writer.WriteInteger(1);

  Writer.WriteString(Font.Name);
  Writer.WriteInteger(Font.Size);
  Writer.WriteInteger(Font.Color);

  Writer.WriteBoolean(FOnlyWholeChars);

{$IFDEF USE_WYSIWYGFONT}
  Writer.WriteBoolean(FUseWYSIWYGFont);
  Writer.WriteInteger(WYSIWYGFont.Font.Angle);
  Writer.WriteInteger(WYSIWYGFont.Font.Color);
  Writer.WriteString(WYSIWYGFont.Font.Name);
  Writer.WriteInteger(Ord(WYSIWYGFont.Font.Pitch));
  Writer.WriteInteger(WYSIWYGFont.Font.Size);
  Writer.WriteInteger(WYSIWYGFont.Font.Width);
{$ENDIF}
end;

procedure TRakStringCell.Assign(Source: TRakCollectionItem);
var
  Src: TRakStringCell;
begin
  inherited Assign(Source);

  if Source is TRakStringCell then
  begin
    Src := TRakStringCell(Source);
    Value := Src.Value;
    FBrush.Assign(Src.FBrush);
  end;
end;

function TRakStringCell.GetAsInteger: integer;
begin
  Result := StrToInt(Value);
end;

procedure TRakStringCell.SetAsInteger(Value: integer);
begin
  AsString := IntToStr(Value);
end;

function TRakStringCell.GetAsString: string;
begin
  Result := FValue;
end;

procedure TRakStringCell.SetAsString(Value: string);
begin
  SetValue(Value);
end;

procedure TRakStringCell.SetValue(Value: string);
begin
  FValue := Value;
  FLastChange := clcData;
  Changed(self);
end;

function TRakStringCell.TextHeight(Zeile: string): integer;
begin
{$IFDEF USE_WYSIWYGFONT}
  if UseWYSIWYGFont then
    Result := WYSIWYGFont.TextHeight(Zeile)
  else
{$ENDIF}
    Result := FCanvas.TextHeight(Zeile);
end;

function TRakStringCell.TextWidth(Zeile: string): integer;
begin
{$IFDEF USE_WYSIWYGFONT}
  if UseWYSIWYGFont then
    Result := WYSIWYGFont.TextWidth(Zeile)
  else
{$ENDIF}
    Result := FCanvas.TextWidth(Zeile);
end;

function  TRakStringCell.GetFontSize: integer;
begin
{$IFDEF USE_WYSIWYGFONT}
  if UseWYSIWYGFont then
    Result := WYSIWYGFont.Font.Size
  else
{$ENDIF}
    Result := FCanvas.Font.Size;
end;

procedure TRakStringCell.SetFontSize(Value: integer);
begin
{$IFDEF USE_WYSIWYGFONT}
  if UseWYSIWYGFont then
    WYSIWYGFont.Font.Size := Value
  else
{$ENDIF}
  begin
    if Assigned(FCanvas) then
      FCanvas.Font.Size := Value;
    Font.Size := Value;
  end;
end;

{$IFDEF USE_WYSIWYGFONT}
function  TRakStringCell.GetFontWidth: integer;
begin
  if UseWYSIWYGFont then
    Result := WYSIWYGFont.Font.Width
  else
    raise Exception.Create('Die Breite kann nur bei einem WYSIWYG-Font ermittelt werden.');
end;

procedure TRakStringCell.SetFontWidth(Value: integer);
begin
  if UseWYSIWYGFont then
    WYSIWYGFont.Font.Width := Value
  else
    raise Exception.Create('Die Breite kann nur bei einem WYSIWYG-Font gesetzt werden.');
end;
{$ENDIF}


constructor TRakStringTable.Create;
begin
  inherited Create;

  FFont := TFont.Create;
  FFont.OnChange := FontChanged;

  FFuncRec := TFuncRec.Create;
  FFuncRec.GetTextWidth := TextWidth;
  FFuncRec.GetTextHeight := TextHeight;
  FFuncRec.GetFontSize := GetFontSize;
  FFuncRec.SetFontSize := SetFontSize;

  FOnlyWholeChars := true;

{$IFDEF USE_WYSIWYGFONT}
  FFuncRec.GetFontWidth := GetFontWidth;
  FFuncRec.SetFontWidth := SetFontWidth;

  FUseWYSIWYGFont := true;
  FWYSIWYGFont := TRakWYSIWYGFont.Create;
{$ENDIF}
end;

destructor TRakStringTable.Destroy;
begin
  FFont.Free;
  FFuncRec.Free;

{$IFDEF USE_WYSIWYGFONT}
  FWYSIWYGFont.Free;
{$ENDIF}

  inherited Destroy;
end;

procedure TRakStringTable.Assign(Source: TRakCustomTableData);
begin
  inherited Assign(Source);

  FFont.Assign(TRakStringTable(Source).FFont);
end;

procedure TRakStringTable.AutoCellsValueHeihgt(Canvas: TCanvas);
begin
  if Rows.Count > 0 then
  begin
    FCanvas := Canvas;

    Canvas.Font.Assign(FFont);
    FitFontSize(FFuncRec, Rows[0].Height);
    FFont.Size := FontSize;
  end;
end;

procedure TRakStringTable.AutoSize(Canvas: TCanvas; Hoehe, Breite: integer);
var
  Row, Col: integer;
begin
  inherited AutoSize(Canvas, Hoehe, Breite);

  // Hier bekommt jede Zelle den Font der Tabelle verpat.
  for Col := 0 to ColCount - 1 do
    for Row := 0 to RowCount - 1 do
      CellsObject[Col, Row].Font.Assign(Font);
end;

procedure TRakStringTable.LoadAsStream(Reader: TReader);
var
  iWert: integer;
begin
  inherited LoadAsStream(Reader);

  iWert := Reader.ReadInteger;
  if iWert > 1 then
    raise TRakTableVersionException.Create('Version 1');


  Font.Name := Reader.ReadString;
  Font.Size := Reader.ReadInteger;
  Font.Color := Reader.ReadInteger;

  FOnlyWholeChars := Reader.ReadBoolean;

{$IFDEF USE_WYSIWYGFONT}
  FUseWYSIWYGFont := Reader.ReadBoolean;
  WYSIWYGFont.Font.Angle := Reader.ReadInteger;
  WYSIWYGFont.Font.Color := Reader.ReadInteger;
  WYSIWYGFont.Font.Name := Reader.ReadString;
  WYSIWYGFont.Font.Pitch := TFontPitch(Reader.ReadInteger);
  WYSIWYGFont.Font.Size := Reader.ReadInteger;
  WYSIWYGFont.Font.Width := Reader.ReadInteger;
{$ENDIF}
end;

procedure TRakStringTable.SaveAsStream(Writer: TWriter);
begin
  inherited SaveAsStream(Writer);

  Writer.WriteInteger(1);

  Writer.WriteString(Font.Name);
  Writer.WriteInteger(Font.Size);
  Writer.WriteInteger(Font.Color);

  Writer.WriteBoolean(FOnlyWholeChars);

{$IFDEF USE_WYSIWYGFONT}
  Writer.WriteBoolean(FUseWYSIWYGFont);
  Writer.WriteInteger(WYSIWYGFont.Font.Angle);
  Writer.WriteInteger(WYSIWYGFont.Font.Color);
  Writer.WriteString(WYSIWYGFont.Font.Name);
  Writer.WriteInteger(Ord(WYSIWYGFont.Font.Pitch));
  Writer.WriteInteger(WYSIWYGFont.Font.Size);
  Writer.WriteInteger(WYSIWYGFont.Font.Width);
{$ENDIF}
end;

procedure TRakStringTable.SetAutoParams;
begin
  inherited SetAutoParams;

{$IFDEF USE_WYSIWYGFONT}
  if UseWYSIWYGFont then
    SetFontWidth(0);
{$ENDIF}
end;

procedure TRakStringTable.SaveAutoParams;
begin
  inherited SaveAutoParams;

{$IFDEF USE_WYSIWYGFONT}
  if UseWYSIWYGFont then
    FSaveFontWidth := GetFontWidth;
{$ENDIF}
end;

procedure TRakStringTable.RestoreAutoParams;
begin
  inherited RestoreAutoParams;

{$IFDEF USE_WYSIWYGFONT}
  if UseWYSIWYGFont then
    SetFontWidth(FSaveFontWidth);
{$ENDIF}
end;

procedure TRakStringTable.FontChanged(Sender: TObject);
var
  Col, Row: integer;
begin
  for Row := 0 to RowCount - 1 do
  begin
    for Col := 0 to ColCount - 1 do
    begin
      CellsObject[Col, Row].Font.Assign(Font);
    end;
  end;
end;

function TRakStringTable.GetAsInteger(Col, Row: integer): integer;
begin
  Result := CellsObject[Col, Row].AsInteger;
end;

procedure TRakStringTable.SetAsInteger(Col, Row: integer; Value: integer);
begin
  CellsObject[Col, Row].AsInteger := Value;
end;

function TRakStringTable.GetCellsObject(Col, Row: integer): TRakStringCell;
begin
  Result := TRakStringCell(inherited CellsObject[Col, Row]);
end;

function TRakStringTable.GetCells(Col, Row: integer): string;
begin
  Result := TRakStringCell(inherited CellsObject[Col, Row]).Value;
end;

procedure TRakStringTable.SetCells(Col, Row: integer; Value: string);
begin
  TRakStringCell(inherited CellsObject[Col, Row]).Value := Value;
end;

function TRakStringTable.GetVisibleCells(Col, Row: integer): string;
begin
  Result := TRakStringCell(inherited CellsObject[Cols.VisibleItems[Col].Index, Row]).Value;
end;

procedure TRakStringTable.SetVisibleCells(Col, Row: integer; Value: string);
begin
  TRakStringCell(inherited CellsObject[Cols.VisibleItems[Col].Index, Row]).Value := Value;
end;

procedure TRakStringTable.SetColCount(Value: integer);
var
  i,t: integer;
  Cell: TRakStringCell;
begin
  for i := 0 to FRows.Count - 1 do
  begin
    if Value > FRows[i].Count then
    begin
      for t := 1 to Value - FRows[i].Count do
      begin
        Cell := TRakStringCell(FRows[i].Add(GetCellType));
        Cell.Font.Assign(FFont);
      end;
    end;
  end;
  inherited SetColCount(Value);
end;

procedure TRakStringTable.SetOnlyWholeChars(Value: boolean);
var
  Row, Col: integer;
begin
  FOnlyWholeChars := Value;

  for Row := 0 to RowCount - 1 do
  begin
     for Col := 0 to ColCount - 1 do
     begin
        CellsObject[Col, Row].OnlyWholeChars := Value;
     end;
  end;
end;

procedure TRakStringTable.SetRowCount(Value: integer);
var
  i,t: integer;
  Row: TRakTableDataRow;
  Cell: TRakStringCell;
begin
  if Value < FRows.Count then
  begin
    for i := FRows.Count -1 downto Value do
      FRows[i].Free;
  end
  else
  begin
    for i := 1 to Value - FRows.Count do
    begin
      Row := FRows.Add;
      for t := 1 to ColCount do
      begin
        Cell := TRakStringCell(Row.Add(GetCellType));
        Cell.Font.Assign(FFont);
        Cell.OnlyWholeChars := OnlyWholeChars;
      end;
    end;
  end;
end;


function TRakStringTable.TextHeight(Zeile: string): integer;
begin
{$IFDEF USE_WYSIWYGFONT}
  if UseWYSIWYGFont then
    Result := WYSIWYGFont.TextHeight(Zeile)
  else
{$ENDIF}
    Result := FCanvas.TextHeight(Zeile);
end;

function TRakStringTable.TextWidth(Zeile: string): integer;
begin
{$IFDEF USE_WYSIWYGFONT}
  if UseWYSIWYGFont then
    Result := WYSIWYGFont.TextWidth(Zeile)
  else
{$ENDIF}
    Result := FCanvas.TextWidth(Zeile);
end;

function  TRakStringTable.GetFontSize: integer;
begin
{$IFDEF USE_WYSIWYGFONT}
  if UseWYSIWYGFont then
    Result := WYSIWYGFont.Font.Size
  else
{$ENDIF}
    Result := FCanvas.Font.Size;
end;

procedure TRakStringTable.SetFontSize(Value: integer);
var
  Col, Row: integer;
begin
  for Row := 0 to RowCount - 1 do
  begin
    for Col := 0 to ColCount - 1 do
      CellsObject[Col, Row].FontSize := Value;
  end;
{$IFDEF USE_WYSIWYGFONT}
  if UseWYSIWYGFont then
    WYSIWYGFont.Font.Size := Value
  else
{$ENDIF}
  begin
    if Assigned(FCanvas) then
      FCanvas.Font.Size := Value;
    Font.Size := Value;
  end;
end;


{$IFDEF USE_WYSIWYGFONT}
function  TRakStringTable.GetFontWidth: integer;
begin
  if UseWYSIWYGFont then
    Result := WYSIWYGFont.Font.Width
  else
    raise Exception.Create('Die Breite kann nur bei einem WYSIWYG-Font ermittelt werden.');
end;

procedure TRakStringTable.SetFontWidth(Value: integer);
var
  Col, Row: integer;
begin
  for Row := 0 to RowCount - 1 do
  begin
    for Col := 0 to ColCount - 1 do
      CellsObject[Col, Row].FontWidth := Value;
  end;
  
  if UseWYSIWYGFont then
    WYSIWYGFont.Font.Width := Value
  else
    raise Exception.Create('Die Breite kann nur bei einem WYSIWYG-Font gesetzt werden.');
end;
{$ENDIF}


function TFuncRec.lGetFontSize: integer;
begin
  Result := FGetFontSize;
end;

procedure TFuncRec.lSetFontSize(Value: integer);
begin
  FSetFontSize(Value);
end;

{$IFDEF USE_WYSIWYGFONT}
function TFuncRec.lGetFontWidth: integer;
begin
  Result := FGetFontWidth;
end;

procedure TFuncRec.lSetFontWidth(Value: integer);
begin
  FSetFontWidth(Value);
end;
{$ENDIF}

function TFuncRec.TextHeight(Zeile: string): integer;
begin
  Result := FGetTextHeight(Zeile);
end;

function TFuncRec.TextWidth(Zeile: string): integer;
begin
  Result := FGetTextWidth(Zeile);
end;





{$IFDEF USE_WYSIWYGFONT}
procedure FitFontWidth(FuncRec: TFuncRec; MaxWidth: integer; Zeile: string);
var
  Min, Max: integer;
  TmpWidth, NextSize: integer;
begin
  if Zeile <> '' then
  begin
    with FuncRec do
    begin
      // Min enthlt die Breite, welche die Schrift mindestens haben mu
      Min := 1;
      // Max enthlt die Breite, bei der die Schrift noch zu breit ist
      Max := -1;
      // Min und Max werden solange aneinander angeglichen, bis in Min
      // die gesuchte Schriftbreite steht. Am Anfang ist Max noch -1, da
      // noch nicht bekannt ist bei welcher Breite die Schrift zu breit wre.
      // Solange noch keine Schriftbreite probiert wurde, die zu breit ist,
      // bleibt Max auf -1 und die nchste Breite, die probiert wird, ist
      // doppelt so gro wie die alte.
      // Sobald Max einen Wert grer als -1 (und auch grer als Min) hat,
      // wird der Mittelwert von Min und Max fr die nchste Breite genommen,
      // die probiert wird.
      // Entsprechend, ob die gerade probierte Breite in den zur Verfgung
      // stehenen Platz pat oder nicht, wird Min oder Max der Wert der
      // gerade probierte Breite gegeben

      // Die erste zu probierende Breite festlegen
      if MaxWidth > 0 then
        FontWidth := MaxWidth
      else
        FontWidth := 1;

      while (Min < Max-1) or (Max = -1) do
      begin // Solange nicht die maximal mgliche Schriftbreite gefunden wurde
        // Die Schriftbreite ermitteln
        TmpWidth := TextWidth(Zeile);

        if TmpWidth < MaxWidth then
        begin // Wenn die Schrift nicht den in der Breite zur Verfgung
              // stehenden Platz ausfllt
          //Merken, da der Font mindestens so breit sein mu, wie er jetzt ist
          Min := FontWidth;
          // Die nchste zu probierende Schriftbreite festlegen
          if Max > -1 then
            NextSize := (Max + Min) div 2
          else
            NextSize := Min * 2;
        end
        else if TmpWidth > MaxWidth then
        begin // Wenn die Schrift den in der Breite zur Verfgung
              // stehenden Platz berschreitet
          //Merken, da der Font nicht so breit sein darf, wie er jetzt ist
          Max := FontWidth;
          // Die nchste zu probierende Schriftbreite festlegen
          NextSize := (Max + Min) div 2;
        end
        else
        begin // Wenn die Schrift den in der Breite zur Verfgung
              // stehenden Platz genau ausfllt
          Min := FontWidth;
          Max := Min;
          NextSize := 1; // Diese Zeile unterdrckt nur eine Warnung
        end;

        if (Min < Max-1) or (Max = -1) then
        begin // Wenn die ideale Schriftbreite noch nicht gefunden wurde
          // Dann wird mit der nchsten weitergemacht
          FontWidth := NextSize;
        end;
      end;

      FontWidth := Min;
    end;
  end;
end;
{$ENDIF}

procedure FitFontSize(FuncRec: TFuncRec; MaxHeight: integer);
var
  Min, Max: integer;
  TmpHeight, NextSize: integer;
begin
  with FuncRec do
  begin
    // Min enthlt die Gre, welche die Schrift mindestens haben mu
    Min := 1;
    // Max enthlt die Gre, bei der die Schrift noch zu gro ist
    Max := -1;
    // Min und Max werden solange aneinander angeglichen, bis in Min
    // die gesuchte Schriftgre steht. Am Anfang ist Max noch -1, da
    // noch nicht bekannt ist bei welcher Gre die Schrift zu Hoch wre.
    // Solange noch keine Schriftgre probiert wurde, die zu gro ist,
    // bleibt Max auf -1 und die nchste Gre, die probiert wird, ist
    // doppelt so gro wie die alte.
    // Sobald Max einen Wert grer als -1 (und auch grer als Min) hat,
    // wird der Mittelwert von Min und Max fr die nchste Gre genommen,
    // die probiert wird.
    // Entsprechend, ob die gerade probierte Gre in den zur Verfgung
    // stehenen Platz pat oder nicht, wird Min oder Max der Wert der
    // gerade probierte Gre gegeben

    // Die erste zu probierende Gre festlegen
    if MaxHeight > 0 then
      FontSize := MaxHeight
    else
      FontSize := 1;

    while (Min < Max-1) or (Max = -1) do
    begin // Solange nicht die maximal mgliche Schriftgre gefunden wurde
      // Die Schriftgre ermitteln
      TmpHeight := TextHeight('JGYPjgyp');

      if TmpHeight < MaxHeight then
      begin // Wenn die Schrift nicht den in der Hhe zur Verfgung
            // stehenden Platz ausfllt
        //Merken, da der Font mindestens so gro sein mu, wie er jetzt ist
        Min := FontSize;
        // Die nchste zu probierende Schriftgre festlegen
        if Max > -1 then
          NextSize := (Max + Min) div 2
        else
          NextSize := Min * 2;
      end
      else if TmpHeight > MaxHeight then
      begin // Wenn die Schrift den in der Hhe zur Verfgung
            // stehenden Platz berschreitet
        //Merken, da der Font nicht so gro sein darf, wie er jetzt ist
        Max := FontSize;
        // Die nchste zu probierende Schriftgre festlegen
        NextSize := (Max + Min) div 2;
      end
      else
      begin // Wenn die Schrift den in der Hhe zur Verfgung
            // stehenden Platz genau ausfllt
        Min := FontSize;
        Max := Min;
        NextSize := 1; // Diese Zeile unterdrckt nur eine Warnung
      end;

      if (Min < Max-1) or (Max = -1) then
      begin // Wenn die ideale Schriftgre noch nicht gefunden wurde
        // Dann wird mit der nchsten weitergemacht
        FontSize := NextSize;
      end;
    end;

    FontSize := Min;
  end;
end;


procedure FitFontSizeByWidth(FuncRec: TFuncRec; MaxWidth: integer; Zeile: string);
var
  Min, Max: integer;
  TmpWidth, NextSize: integer;
begin
  if Zeile <> '' then
  begin
    with FuncRec do
    begin
      // Min enthlt die Gre, welche die Schrift mindestens haben mu
      Min := 1;
      // Max enthlt die Gre, bei der die Schrift noch zu gro ist
      Max := -1;
      // Min und Max werden solange aneinander angeglichen, bis in Min
      // die gesuchte Schriftgre steht. Am Anfang ist Max noch -1, da
      // noch nicht bekannt ist bei welcher Gre die Schrift zu Hoch wre.
      // Solange noch keine Schriftgre probiert wurde, die zu gro ist,
      // bleibt Max auf -1 und die nchste Gre, die probiert wird, ist
      // doppelt so gro wie die alte.
      // Sobald Max einen Wert grer als -1 (und auch grer als Min) hat,
      // wird der Mittelwert von Min und Max fr die nchste Gre genommen,
      // die probiert wird.
      // Entsprechend, ob die gerade probierte Gre in den zur Verfgung
      // stehenen Platz pat oder nicht, wird Min oder Max der Wert der
      // gerade probierte Gre gegeben

      // Die erste zu probierende Gre festlegen
      if FontSize = 0 then
        FontSize := 16;

      while (Min < Max-1) or (Max = -1) do
      begin // Solange nicht die maximal mgliche Schriftgre gefunden wurde
        // Die Schriftgre ermitteln
        TmpWidth := TextWidth(Zeile);

        if TmpWidth < MaxWidth then
        begin // Wenn die Schrift nicht den in der Hhe zur Verfgung
              // stehenden Platz ausfllt
          //Merken, da der Font mindestens so gro sein mu, wie er jetzt ist
          Min := FontSize;
          // Die nchste zu probierende Schriftgre festlegen
          if Max > -1 then
            NextSize := (Max + Min) div 2
          else
            NextSize := Min * 2;
        end
        else if TmpWidth > MaxWidth then
        begin // Wenn die Schrift den in der Hhe zur Verfgung
              // stehenden Platz berschreitet
          //Merken, da der Font nicht so gro sein darf, wie er jetzt ist
          Max := FontSize;
          // Die nchste zu probierende Schriftgre festlegen
          NextSize := (Max + Min) div 2;
        end
        else
        begin // Wenn die Schrift den in der Hhe zur Verfgung
              // stehenden Platz genau ausfllt
          Min := FontSize;
          Max := Min;
          NextSize := 1; // Diese Zeile unterdrckt nur eine Warnung
        end;

        if (Min < Max-1) or (Max = -1) then
        begin // Wenn die ideale Schriftgre noch nicht gefunden wurde
          // Dann wird mit der nchsten weitergemacht
          FontSize := NextSize;
        end;
      end;

      FontSize := Min;
    end;
  end;
end;

procedure TRakCustomTableData.SetAutoFormatType(const Value: TAutoFormatType);
begin
  FFAutoFormatType := Value;
end;

procedure TRakCustomTableData.NormalizeCellSize;
begin

end;

procedure TRakStringTable.NormalizeCellSize;
var
  Row, Col: integer;
begin
  for Row := 0 to RowCount - 1 do
  begin
    for Col := 0 to ColCount -1 do
    begin
      CellsObject[Col, Row].SetFontSize(FontSize);
{$IFDEF USE_WYSIWYGFONT}
      CellsObject[Col, Row].SetFontWidth(GetFontWidth);
{$ENDIF}
    end;
  end;
end;

procedure TRakCustomTableData.SetDrehung(const Value: integer);
begin
  FDrehung := Value;
end;

function TRakStringTable.GetCellType: TRakTableCellClass;
begin
  Result := TRakStringCell;
end;

{$IFDEF USE_WYSIWYGFONT}
procedure TRakStringCell.SetDoClip(const Value: boolean);
begin
  FDoClip := Value;
end;
{$ENDIF}

end.
